/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __Quaternion_hh__
#define __Quaternion_hh__

#include <Base/Math/Vector4.hh>
#include <Base/Math/Vector3.hh>
#include <Base/Geometry/RMatrixBase.hh>
#include <float.h>

namespace BIAS {

#define QUATERNION_EPSILON DBL_EPSILON
#define QUATERNION_TYPE double

  // forward declaration to avoid mutual inclusion
  class RMatrixBase;
  /** @class Quaternion
      @ingroup g_geometry
      @brief class for rotation with axis and angle

      Quaternions can be used for rotation around an axis through the
      origin by spcecifing the normal direction vector of the axis and
      an angle.
     
      The rotation for positive angles is clockwise,        
      when looking in direction of the axis.

      Contains qx=[0], qy=[1], qz=[2], qw=[3].

      Not every quaternion describes a rotation.
      Only quaternions of the form:     
    
      qx = x * sin(phi/2) , which is the imaginary part i
     
      qy = y * sin(phi/2)  , which is the imaginary part j
     
      qz = z * sin(phi/2)  , which is the imaginary part k
     
      qw = cos(phi/2),     , which is the real part    

      where (x, y, z) is the normalized direction vector of the axis 
      and phi is the angle of rtoation.
     

      @author grest 06 2003 */
  template<class QUAT_TYPE> 
  class DLLEXPORTAPI Quaternion : public Vector4<QUAT_TYPE>
  {
   
  public:
    inline Quaternion() : Vector4<QUAT_TYPE>() {};

    /**	copy constructor
        @author Daniel Grest (06/2003) 
    */
    inline Quaternion(const Quaternion<QUAT_TYPE> &q) { *this = q; };        
    
    Quaternion(QUAT_TYPE i, QUAT_TYPE j, QUAT_TYPE k,
               QUAT_TYPE r);
  
    inline ~Quaternion() {};
  
    inline void SetIdentity();

    /** returns the Inverse rotation Quaternion
     */
    inline Quaternion<QUAT_TYPE> Inverse() const;
    
    /** inverts this, by changing the rotation axis by *=-1 
     */
    inline void Invert();

    /** quaternion multiplication: this= this * quat
     */
    inline void Mult(Quaternion<QUAT_TYPE> &quat);

    /** quaternion multiplication: this = quat * this
     */
    inline void MultLeft(Quaternion<QUAT_TYPE> &quat);
  
    /** rotates the given Vector qith the quaternion ( q v q* )
        the resulting vector is given in res
        @returns 0 in case of no error
        @author Daniel Grest, June 2003
    */
    inline int MultVec(Vector3<QUAT_TYPE> &vec, 
                       Vector3<QUAT_TYPE> &res);

    /** sets the quaternion with given rotation axis and angle (in rad)
        @returns 0 in case of no error
        @author Daniel Grest, June 2003
    */
    inline void SetValueAsAxisRad(const Vector3<QUAT_TYPE> &axis,
                                  QUAT_TYPE angle);
 
    /** sets the quaternion with given rotation axis and angle (in rad)
        @returns 0 in case of no error
        @author Daniel Grest, June 2003
    */

    inline void SetValueAsAxisRad(QUAT_TYPE axisX, QUAT_TYPE axisY,
                                  QUAT_TYPE axisZ, QUAT_TYPE angle);
 
    /** returns a corresponding rotation matrix 
        @author woelk 01 2003 untested */
    int GetRotationMatrix(RMatrixBase& R);
  
    /** returns rotation in axis and angle notation, angle in rad
        @author woelk 12 2002 */
    int GetAxisAngle(Vector3<QUAT_TYPE>& axis, QUAT_TYPE& angle);


    /** scale vector4 so that norm = 1        
        The use of this is questionable...  who wrote this? (grest)
    */
    inline void Normalize() {
      (*this) /= (QUAT_TYPE)NormL2();
    }


    /** assignment operator
        @author grest 06 2003 
    */
    inline Quaternion<QUAT_TYPE>& operator=(const Quaternion<QUAT_TYPE>& vec) {
      memcpy((void *)this->data_, (void *)vec.GetData(), 
             VECTOR4SIZE*sizeof(QUAT_TYPE)); 
      return *this; 
    };  
    
  }; // class

  //#include "Base/Geometry/Quaternion.inl"
#include "QuaternionInl.hh"


} // namespace

#endif // __Quaternion_hh__
