/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _BIAS_IMAGE_DRAW_H_
#define _BIAS_IMAGE_DRAW_H_
#include "bias_config.h" 

#include "Image.hh"

namespace BIAS {
  /** @class ImageDraw
      @ingroup g_image_other
      @brief drawing simple entities into the image like rectangles or lines
      
      
      As all fucntions are static they have to be called by
      eg. ImageDraw::RectangleCenter() and not with an object.
  */
  template <class StorageType>
  class ImageDraw 
  {
    /** class for Drawing in Images
        @author fw */
  public:
    /* returns value with contrast to t */
    static StorageType GetContrastValue(StorageType t);

    ////////////////////////////////////////////////////////////
    /// rectangles
    ////////////////////////////////////////////////////////////

    ////////////////////////////////////
    // specifying rectangle corners

    /** Draws the rectangle defined by upper left and bottom right
        corner with value[i] in channel i **/
    static int RectangleCorners(Image<StorageType>& im, unsigned int minx, 
                         unsigned int miny, unsigned int maxx, 
                         unsigned int maxy, StorageType value[]);

    /** as above but fills all channels with value */
    static int RectangleCorners(Image<StorageType>& im, unsigned int minx, 
                         unsigned int miny, unsigned int maxx, 
                         unsigned int maxy, StorageType value);

    /** decides which color to use */
    static int RectangleCorners(Image<StorageType>& im, unsigned int minx, 
                         unsigned int miny, unsigned int maxx, 
                         unsigned int maxy);

    /** faster for one channel images */
    static int RectangleCornersGrey(Image<StorageType>& im, unsigned int minx, 
                             unsigned int miny, unsigned int maxx, 
                             unsigned int maxy, StorageType GreyValue);

    /** faster for one channel images */
    static int RectangleCornersGreyFill(Image<StorageType>& im, 
                                        unsigned int minx, unsigned int miny, 
                                        unsigned int maxx, unsigned int maxy, 
                                        StorageType GreyValue);
    //////////////////////////
    // specifying the rectangle center and size

    /** Draws the rectangle around X, Y with Size and Value[i] in channel i
        @author Felix Woelk **/
    static int RectangleCenter(Image<StorageType>& im,
                               unsigned int x, unsigned int y, 
                               unsigned int size, StorageType value[]);

    /** fills the channels with value */
    static int RectangleCenterGrey(Image<StorageType>& im,
                             unsigned int x, unsigned int y, 
                             unsigned int size, StorageType value);

    static int RectangleCenterGreyFill(Image<StorageType>& im,
                                       unsigned int x, unsigned int y, 
                                       unsigned int size, StorageType value);

    /** convenience wrapper */
    inline 
    static int RectangleCenter(Image<StorageType>& im, unsigned int coo[2], 
                         unsigned int size, StorageType value[])
    { return RectangleCenter(im, coo[0], coo[1], size, value); };

    /** draws a rectangle of size around x and y, decides which color to use */
    static int RectangleCenter(Image<StorageType>& im, unsigned int x, 
                         unsigned int y, unsigned int size);

    /** convenience wrapper */
    static int RectangleCenter(Image<StorageType>& im, unsigned int coo[2], 
                         unsigned int size);


    //////////////////////////////////////////////////////////////
    /// lines
    /////////////////////////////////////////////////////////////


    /** Draws a line between the two points using the bresenham algorithm
        uses value[i] in channel i **/
    static int Line(Image<StorageType>& im, const unsigned int start[2],
                    const unsigned int end[2], StorageType value[]);

    /** uses value in every channel */
    static int Line(Image<StorageType>& im, const unsigned int start[2],
                    const unsigned int end[2], StorageType value);

    /** faster algorithm for 1 channel images */
    static int LineGrey(Image<StorageType>& im, const unsigned int start[2],
                        const unsigned int end[2], StorageType value);
    
    /** decides which color to use */
    static int Line(Image<StorageType>& im, const unsigned int start[2],
                    const unsigned int end[2]);

    /** faster algorithm for 1 channel images, decides which color to use */
    static int LineGrey(Image<StorageType>& im, const unsigned int start[2],
                        const unsigned int end[2]);

    //////////////////////////
    // just wrappers

    /** convenience wrapper */
    static int Line(Image<StorageType>& im, const unsigned int StartX, 
                    const unsigned int StartY, const unsigned int EndX, 
                    const unsigned int EndY, StorageType Value[]);

    /** convenience wrapper */
    static int Line(Image<StorageType>& im, const unsigned int StartX, 
                    const unsigned int StartY, const unsigned int EndX, 
                    const unsigned int EndY, StorageType Value);

    /** convenience wrapper */
    static int Line(Image<StorageType>& im, const unsigned int StartX, 
                    const unsigned int StartY, const unsigned int EndX, 
                    const unsigned int EndY);

    /** convenience wrapper */
    static int LineGrey(Image<StorageType>& im, const unsigned int StartX, 
                        const unsigned int StartY, const unsigned int EndX, 
                        const unsigned int EndY, StorageType Value);

    
    inline static int Line(Image<StorageType>& im, unsigned int coo[4],
                           StorageType value[])
    { return Line(im, &coo[0], &coo[2], value); };

    


    ///////////////////////////
    // interpolated lines

    /** draws an interpolated line 
        just a wrapper for the below
        @author Felix Woelk */
    static int InterpolatedLine(Image<StorageType>& im, 
                                const int StartX, const int StartY,
                                const int EndX, const int EndY,
                                StorageType Value[]);

    static int InterpolatedLineGrey(Image<StorageType>& im, 
                                    const int StartX, const int StartY,
                                    const int EndX, const int EndY,
                                    StorageType Value);

    inline 
    static int InterpolatedLineGrey(Image<StorageType>& im, 
                                    unsigned int coo[4],
                                    StorageType Value)
    { return InterpolatedLineGrey(im, coo[0], coo[1], coo[2], 
                                      coo[3], Value); };


    ////////////////////////////////////
    // arrows

    /** draws an arrow from start to end, the tips of the head are 
        length pixel back on the line and width pixel away from the line
        @author woelk 12/2003 */
    static int Arrow(Image<StorageType>& im, const unsigned start[2],
                     const unsigned end[2], const unsigned length,
                     const unsigned width, StorageType value[]);
    


    /////////////////////////////////////////
    // circles

    static int CircleCenter(Image<StorageType>& im, unsigned int CenterX,
                            unsigned int CenterY, unsigned int Radius, 
                            StorageType Value[]);

    static int CircleCenterFilled(Image<StorageType>& im, unsigned int CenterX,
                                  unsigned int CenterY, unsigned int Radius, 
                                  StorageType Value[]);

    /* TODO
       static int InterpolatedLineColor(Image<StorageType>& im, 
       const int StartX, const int StartY,
       const int EndX, const int EndY,
       StorageType Value[]); */

  }; // class ImageDraw

}  // namespace

#endif // _BIAS_IMAGE_DRAW_H_
