/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIASUUID_hh__
#define __BIASUUID_hh__

#include <ctime>
#include <Base/Debug/Error.hh>
#include <Base/Debug/Debug.hh>

#ifdef WIN32
// on Win32 there are two types: GUID and UUID, both pointing to the same struct
  #include <rpc.h>
//  #include <rpcdce.h>
// remove UUID to avoid name clash with our class
  #undef UUID
  #undef uuid_t
  typedef GUID uuid_t;  // let our internal data type be of type GUID
  #pragma warning (disable: 4238)
#else
  #include <uuid/uuid.h>
#endif

#include <time.h>

namespace BIAS {

  /** @class UUID     
      @ingroup g_utils
      @brief interface class for 
      producing/storing Universally Unique IDentifiers

       A UUID is a 128 bit value computed from the MAC address of the network 
       adapter (which means "spatial uniqueness") and a high resolution 
       timestamp providing "temporal uniqueness"

       @attention temporal uniqueness is only obtained if time is monotonic,
       i.e. you MUST NOT turn back the clock (-> NTP ?)
           
       @author koeser 09/2003 */

  class DLLEXPORTAPI UUID : public Debug
  {
  public:

    /** @brief static function which simply produces a uuid and returns 
        @attention calling this function on an uninitialized UUID object does
                   not initialize it. It is _really_ a static function. 
        @return a new uuid which can be assigned to a UUID object */
    static UUID GenerateUUID();
 
    /** @brief default constructor, creates empty uuid */
    UUID() { IsValid_ = false; };  
    
    /** @brief use this if you want to create a UUID "manually" */
    explicit UUID(const uuid_t &id) { 
      memcpy((void*)&theID_,(void*)&id,sizeof(uuid_t));
      IsValid_ = true;
    };

    /** @brief default copy constructor, only calls operator= */
    UUID(const BIAS::UUID& id) { (*this) = id; };

    /** @brief construct from string containing ascii uuid */
    explicit UUID(const std::string &sID);

    /** @brief construct from string containing ascii uuid 
        @return true if succesful */
    bool SetFromString(const std::string &sID);
    
    /** @brief returns the time used to create the uuid */
    time_t GetUUIDCreationTime() const;

    /** @brief comparison operator only compares ids, no check of validities */
    inline bool operator==(const UUID& id) const { 
#ifdef _WIN32
      RPC_STATUS dummy;
      return (!UuidCompare(&((GUID)id.theID_),&((GUID)theID_),&dummy));
#else
      return (!uuid_compare(id.theID_,theID_));
#endif
    };

    /** @brief "not equal"-operator, uses result of operator== */
    inline bool operator!=(const UUID& id) const { 
      return (!((*this)==id)); };

    /** @brief the operator< does not express temporal ordering, it does only
        compare the data areas of the two ids, e.g. for sorting, map-storage */
    inline bool operator<(const UUID& id) const { 
#ifdef _WIN32
      RPC_STATUS dummy;
      return (UuidCompare(&((GUID)id.theID_),&((GUID)theID_),&dummy)<0);
#else
      return (uuid_compare(id.theID_,theID_)<0);
#endif
    };

    /** @brief operator<= does not express temporal ordering, it does only
        compare the data areas of the two ids !!! */
    inline bool operator<=(const UUID& id) const { 
#ifdef _WIN32
      RPC_STATUS dummy;
      return (UuidCompare(&((GUID)id.theID_),&((GUID)theID_),&dummy)<=0);
#else
      return (uuid_compare(id.theID_,theID_)<=0);
#endif
    };

    /** @brief standard assignment operator */
    inline UUID& operator=(const UUID& id) { 
      memcpy((void*)&theID_,(void*)&id.theID_,sizeof(uuid_t)); 
      IsValid_ = id.IsValid_; return (*this);};

    /** @brief writes the UUID into a string object */
    void GetString(std::string& sUUID) const;

    /** @brief checks whether this uuid is valid(true) or unitialized(false) */
    inline bool IsValid() const { return IsValid_; };

    /** @brief sets the valid flag to false */
    inline void Invalidate() { IsValid_ = false; };


  protected:
    /// when creating a uuid, current time is compared against the time
    /// saved during the last uuid generation
	///
	/// Hint: If you have WIN32 compilation problems look for WIN32_LEAN_AND_MEAN and windows.h (JW) 
    static timeval LastGenerationTime_;

    /// the data fields of the uuid
    uuid_t theID_;

    /// flag indicating whether this uuid has been assigned a value yet
    bool IsValid_;
  
    /** @relates UUID
        @brief write a UUID to a stream */
    inline friend std::ostream& operator<<(std::ostream & os, const UUID& id) {
       std::string suid;
    id.GetString(suid); 
       os << suid;
      return os; 
    }

    /** @relates UUID
        @brief reads a UUID from a stream */
    inline friend std::istream& operator>>(std::istream & is, UUID& id) {
      char cUUID[37];
      is.get(cUUID, 37);
      cUUID[36] = 0;
      id.IsValid_ = true;
      // call libuuid function to get id out of string:
#ifdef _WIN32
      unsigned char * cUUIDwin = (unsigned char *)cUUID;
      if (UuidFromString(cUUIDwin, &((GUID)id.theID_))!=RPC_S_OK) {
        BIASERR("UUID could not be read from <"<<cUUID<<">");
        id.IsValid_ = false;
      }
#else
      if (uuid_parse(cUUID, id.theID_)!=0) {
        BIASERR("UUID could not be read from <"<<cUUID<<">");
        id.IsValid_ = false;
      }
#endif
      return is;
    }
    
  }; // class
} // namespace

#endif // __BIASUUID_hh__
