/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Matrix3x3.hh"

using namespace BIAS;
using namespace std;


template<class T>
Matrix3x3<T>::Matrix3x3<T>(const Matrix<T>& mat) : Matrix<T>(mat)
{
  if ((mat.num_rows() != 3) || (mat.num_cols() != 3)) {
    BIASERR("wrong matrix size "<<mat.num_rows()<<"x"
           << mat.num_cols());
  }
}

template<class T>
Matrix3x3<T>::Matrix3x3<T>(const TNT::Matrix<T>& mat) : Matrix<T>(mat)
{
  if ((mat.num_rows() != 3) || (mat.num_cols() != 3)) {
    BIASERR("wrong matrix size "<<mat.num_rows()<<"x"
           << mat.num_cols());
  }
}

template<class T>
Matrix3x3<T>::~Matrix3x3() 
{}


template<class T> 
Matrix3x3<T> Matrix3x3<T>::Transpose() const
{
    Matrix3x3<T> ret;
    ret = *this;

    T *retd = ret.GetData();
    T *thisd = this->GetData();
    retd[1] = thisd[3];
    retd[3] = thisd[1];
    retd[2] = thisd[6];
    retd[6] = thisd[2];
    retd[5] = thisd[7];
    retd[7] = thisd[5];
    return ret;
}



template <class T> 
Matrix3x3<T>& Matrix3x3<T>::newsize(int rows, int cols) 
{
  if ((rows != 3) || (cols != 3)) {
    // do nothing
    // print error message if the new size should not be a 3x4 matrix
    BIASERR("The size of a Matrix3x3<T> has to be 3 rows, 3 cols. "
           <<"newsize can't create a " 
           << rows << " x " << cols << " Matrix3x3 !");
#if DEBUG
    exit(-1);
#endif
    return *this;
  } else {
    BIASASSERT( rows==3 ); // 3x3 matrix
    BIASASSERT( cols==3 );
    // call base class function 
    // (usually unnecessary except for first construction)
    Matrix<T>::newsize(3, 3);
    return *this;
  }
}

template<class T>
void Matrix3x3<T>::SetFromVector(const TNT::Vector<T>& vec) 
{
  BIASASSERT( vec.size() == 9 );
  // fill the matrix rowwise with the first 9 elements of the vector
  // example:
  // 1 2 3 4 5 6 7 8 9
  //  -->
  // 1 2 3
  // 4 5 6
  // 7 8 9
  for (int i=0; i<9; i++) {
    this->GetData()[i] = vec[i];
  }
}

//JW 07/2003
template<class T>
void Matrix3x3<T>::SetFromColumnVectors(const BIAS::Vector3<T> & v0,
                                        const BIAS::Vector3<T> & v1,
                                        const BIAS::Vector3<T> & v2 )
{
  // will be loop-unrolled ..
  // for all elements of a column:
  for (unsigned int row=0; row<3; row++){
    (*this)[row][0]=v0[row];
    (*this)[row][1]=v1[row];
    (*this)[row][2]=v2[row];
  };
}

//JW 07/2003
template<class T>
void Matrix3x3<T>::SetFromRowVectors(const BIAS::Vector3<T> & v0,
                                     const BIAS::Vector3<T> & v1,
                                     const BIAS::Vector3<T> & v2 )
{
  // will be loop-unrolled ..
  // for all elements of a row:
  for (unsigned int col=0; col<3; col++){
    (*this)[0][col]=v0[col];
    (*this)[1][col]=v1[col];
    (*this)[2][col]=v2[col];
  };
}

// JW
template<class T>
void Matrix3x3<T>::GetRow(const unsigned int row, Vector3<T>& v) const 
{
  BIASASSERT(row<3);// this has only 3 rows[0,1,2]
  T* p=this->GetData()+3*row;
  v[0]=*p++;
  v[1]=*p++;
  v[2]=*p;
  //  BIAS::Vector3<T> v( &(*this)[row][0] );
  //return v;
}

// JW
template<class T>
BIAS::Vector3<T> Matrix3x3<T>::GetColumn(const unsigned int col) const {
  BIASASSERT(col<3);// this has only 3 cols[0,1,2]
  BIAS::Vector3<T> v( (*this)[0][col], (*this)[1][col], (*this)[2][col] );
  return v;
}

template<class T>
T Matrix3x3<T>::GetDeterminant()
{
  double t4 = (*this)[0][0]*(*this)[1][1];
  double t6 = (*this)[0][0]*(*this)[1][2];
  double t8 = (*this)[0][1]*(*this)[1][0];
  double t10 = (*this)[0][2]*(*this)[1][0];
  double t12 = (*this)[0][1]*(*this)[2][0];
  double t14 = (*this)[0][2]*(*this)[2][0];
  return T((t4*(*this)[2][2]-t6*(*this)[2][1]-t8*(*this)[2][2]+
          t10*(*this)[2][1]+t12*(*this)[1][2]-t14*(*this)[1][1]));
}

template<class T>
Matrix3x3<T> Matrix3x3<T>::Invert() const
{
  Matrix3x3<T> x;
  // following lines have been produced by Maple  
  double t4 = (*this)[0][0]*(*this)[1][1];
  double t6 = (*this)[0][0]*(*this)[1][2];
  double t8 = (*this)[0][1]*(*this)[1][0];
  double t10 = (*this)[0][2]*(*this)[1][0];
  double t12 = (*this)[0][1]*(*this)[2][0];
  double t14 = (*this)[0][2]*(*this)[2][0];
  // matrices with zero determinante ar not invertible
  BIASASSERT((t4*(*this)[2][2]-t6*(*this)[2][1]-t8*(*this)[2][2]+t10
              *(*this)[2][1]+t12*(*this)[1][2]-t14*(*this)[1][1])!=0.0);
  double t17 = 1.0/(t4*(*this)[2][2]-t6*(*this)[2][1]-t8*(*this)[2][2]+t10
                    *(*this)[2][1]+t12*(*this)[1][2]-t14*(*this)[1][1]);

  x[0][0] = T(((*this)[1][1]*(*this)[2][2]-(*this)[1][2]*(*this)[2][1])*t17);
  x[0][1] = T(-((*this)[0][1]*(*this)[2][2]-(*this)[0][2]*(*this)[2][1])*t17);
  x[0][2] = T(-(-(*this)[0][1]*(*this)[1][2]+(*this)[0][2]*(*this)[1][1])*t17);
  x[1][0] = T(-((*this)[1][0]*(*this)[2][2]-(*this)[1][2]*(*this)[2][0])*t17);
  x[1][1] = T(((*this)[0][0]*(*this)[2][2]-t14)*t17);
  x[1][2] = T(-(t6-t10)*t17);
  x[2][0] = T(((*this)[1][0]*(*this)[2][1]-(*this)[1][1]*(*this)[2][0])*t17);
  x[2][1] = T(-((*this)[0][0]*(*this)[2][1]-t12)*t17);
  x[2][2] = T((t4-t8)*t17);

  return x;
}


