/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _Matrix3x4_hh_
#define _Matrix3x4_hh_

#include "bias_config.h" 

#include <Base/Debug/Error.hh>
  //#include <Base/Debug/Debug.hh>

#include "Matrix.hh"
#include "Operators.hh"


namespace BIAS {

  template <class T> class Matrix3x3;
/**
 
    @class Matrix3x4
    @ingroup g_math
    @brief is a 'fixed size' rectangular matrix of dim. 3 x 4 
    which is templated over the element-type
    
    @author Jan Woetzel
    @status untested (04/17/2002)
**/

  // forward declaration to avoid mutual inclusion
  template <class T> class Vector4;

  template<class T=double>
  class DLLEXPORTAPI Matrix3x4 : public Matrix<T> {

  public:
    /**	default constructor
        @status untested **/
    Matrix3x4<T>() : Matrix<T>(3, 4) {};


    /** @author Jan Woetzel
        @status untested (04/17/2002) **/
    explicit Matrix3x4<T>(const char *s) : Matrix<T>(3, 4, s) {};


    /** @author Jan Woetzel
        @status untested (04/17/2002) **/
    Matrix3x4<T>(const Matrix3x4<T> & A) : Matrix<T>(A) {};


    /** destructor
        @status untested (04/17/2002)**/
    virtual ~Matrix3x4<T>();


    /** matrix - vector multiplicate this matrix with Vector4, 
        storing the result in destvec
        calculates:
        destvec = (this Matrix) * argvec
        @author  Jan Woetzel
        @status untested (04/17/2002) **/
    inline void Mult(const Vector4<T> &argvec, Vector3<T> &destvec) const;


    /** assignment operator
        calling corresponding operator from base class if appropriate.
        @author Jan Woetzel
        @status untested (02/25/2002) **/
    Matrix3x4 & operator= (const Matrix<T> &mat);


    /** set the elements of this matrix to the matrix
        1  0  0  0
        0  1  0  0
        0  0  1  0
        which is no strict identity (!)
        @author Jan Woetzel
        @date 04/17/2002
        @status untested **/
    inline void set_identity();


    /** just neccessary to avoid resizing of this 'fixed size' matrix 
        because it is derived from teh resizable Matrix
        Should be removed if Matrix3x4 becomes a base class.
        @woetzel (04/17/2002) 
        @author Jan Woetzel
        @vollpfosten for unnecessary work **/
    Matrix3x4 & newsize(int rows, int cols);


  }; // class Matrix3x4
  

  template <class T> inline void 
  Matrix3x4<T>::Mult( const Vector4<T> &argvec, Vector3<T> &destvec) const
  {
    // direct access to the matrix values using fixed indicies:
    //  0  1  2  3       0
    //  4  5  6  7   *   1
    //  8  9 10 11       2
    //                   3
    const register T *matP = this->GetData();	
    // first row:
    destvec[0] =
      matP[0] * argvec[0]
      +matP[1] * argvec[1]
      +matP[2] * argvec[2]
      +matP[3] * argvec[3];
    // second row:
    destvec[1] =
      matP[4] * argvec[0]
      +matP[5] * argvec[1]
      +matP[6] * argvec[2]
      +matP[7] * argvec[3];
    // third row (because of 3x4 there are only 3 rows):
    destvec[2] =
      matP[8]  * argvec[0]
      +matP[9]  * argvec[1]
      +matP[10] * argvec[2]
      +matP[11] * argvec[3];
  }


  template <class T> inline void 
  Matrix3x4<T>::set_identity() 
  {
    // set this matrix to teh following (which is no strict identity matrix)
    //  1  0  0  0
    //  0  1  0  0
    //  0  0  1  0
    register T* d = this->GetData(); // pointer to this matrix's data array
    d[0] = 1;
    d[1] = 0;
    d[2]  = 0;
    d[3] = 0;
    d[4] = 0;
    d[5] = 1;
    d[6]  = 0;
    d[7] = 0;
    d[8] = 0;
    d[9] = 0;
    d[10] = 1;
    d[11] = 0;
  }

} // namespace BIAS

#endif

