/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Gauss.hh"

using namespace BIAS;
using namespace std;


//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////
#define FILTER_INSTANTIATION_CLASS Gauss
#include "Filterinst.hh"


//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////


template <class InputStorageType, class OutputStorageType>
Gauss<InputStorageType, OutputStorageType>::
Gauss()
  : Convolution<InputStorageType, OutputStorageType>() 
{
  // no param object !
  this->_ParamsAllocated = true;
  _GaussSigma=new double;
  *_GaussSigma = 0.7;
  _GaussRatio=new double;
  *_GaussRatio = 0.01;
  _LastRatio = _LastSigma = -1.0;
}

template <class InputStorageType, class OutputStorageType>
Gauss<InputStorageType, OutputStorageType>::
Gauss(Param& para, const std::string& postfix)
  : Convolution<InputStorageType, OutputStorageType>(para, "-Gauss"+postfix) 
{
  _LastRatio = _LastSigma = -1.0;
  int grp = para.GetFreeGroupID();
  _AddParam(para, grp, "-Gauss"+postfix);
  if (para.IsUsedGroupID(grp)) para.SetGroupName(grp, "Gauss");
}

template <class InputStorageType, class OutputStorageType>
Gauss<InputStorageType, OutputStorageType>::
Gauss(Param& para, const int grp, const std::string& postfix)
  : Convolution<InputStorageType,OutputStorageType>(para,grp, "-Gauss"+postfix) 
{
  _LastRatio = _LastSigma = -1.0;
  _AddParam(para, grp, "-Gauss"+postfix);
}

template <class StorageType, class KernelType>
Gauss<StorageType, KernelType>::~Gauss()
{
  if (this->_ParamsAllocated) {
    delete _GaussSigma; _GaussSigma = NULL;
    delete _GaussRatio; _GaussRatio = NULL;
  }
}


////////////////////////////////////////////////////////////////////////////

template <class InputStorageType, class OutputStorageType>
int Gauss<InputStorageType, OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& dst)
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Gauss::Filter\n");
  if ((*_GaussSigma != _LastSigma)||(*_GaussRatio != _LastRatio))
    _CalculateKernels(*_GaussSigma, *_GaussRatio);
  return Convolution<InputStorageType, OutputStorageType>::Filter(src,dst);
}

template <class InputStorageType, class OutputStorageType>
int Gauss<InputStorageType, OutputStorageType>::
Filter3x3(const Image<InputStorageType>& src, Image<OutputStorageType>& dst) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Gauss::Filter3x3\n");
  
  BIASWARN("Fast Gauss Filter3x3 not implemented calling Filter.");
  return Filter(src, dst);
}
  
template <class InputStorageType, class OutputStorageType>
int Gauss<InputStorageType, OutputStorageType>::
Filter5x5(const Image<InputStorageType>& src, Image<OutputStorageType>& dst) 
{
  BIASCDOUT(D_FILTERBASE_CALLSTACK, "Gauss::Filter5x5\n");
  BIASWARN("Fast Gauss Filter5x5 not implemented calling Filter.");
  return Filter(src, dst);
}
 
////////////////////////////////////////////////////////////////////////////

template <class InputStorageType, class OutputStorageType>
void Gauss<InputStorageType, OutputStorageType>::
_CalculateKernels(double Sigma, double Ratio)
{
  const int hws = (int)(floor(sqrt(-2.0*Sigma*Sigma*log(Ratio))));
  const int size = (hws<<1)+1;
  
  Vector<FM_FLOAT> H(size), V(size);
  

  BIASCDOUT(D_CONV_KERNEL, "Sigma = "<<Sigma<<"\tRatio = "<<Ratio<<endl);
  BIASCDOUT(D_CONV_KERNEL, "size = "<<size<<"\thws = "<<hws<<endl);

  int i, ip, im;
  double fac=1.0/(2.0*Sigma*Sigma);
  register double g;
  double sum=0.0;
  // calculate the filter coefficients as gauss function
  for (i=0; i<=hws; i++){
    ip = hws+i;
    im = hws-i;
    g  = expf((float)(-(double)(i*i)*fac));
    V[ip] = V[im] = H[im] = H[ip] = (FM_FLOAT)g;
    //BIASCDOUT(D_CONV_KERNEL, i<<"v:"<<this->_VVec[im]<<"  "<<endl);
    sum += g*2.0;
  }
  sum -= V[hws]; // added twice
  BIASCDOUT(D_CONV_KERNEL, "sum = "<<sum<<endl);

  // normalize filter, so that the sum is 1.0
  sum = 1.0/sum;
  BIASCDOUT(D_CONV_KERNEL, "sum = "<<sum<<endl);
  for (i=0; i<=hws; i++){
    ip=hws+i;
    im=hws-i;
    H[ip] = H[im] = V[ip] = V[im] = V[im] * (FM_FLOAT)sum;
    BIASCDOUT(D_CONV_KERNEL, H[im]<<"  "<<endl);
  }
  _fm.Init(H,V);
  // compute the int approximation
  int thebits = _fm.ComputeIntPrecisionBits(sizeof(InputStorageType)*8,
                                            sizeof(CONV_INT)*8-1);
  _fm.CreateIntFilter(thebits,thebits,thebits);

  _LastSigma = Sigma;
  _LastRatio = Ratio;

#ifdef BIAS_DEBUG
  if (DebugLevelIsSet(D_CONV_KERNEL)){
    sum = 0.0;
    for (i=-hws; i<=hws; i++){
      sum += H[i+hws];
      BIASCDOUT(D_CONV_KERNEL, H[i+hws]<<"  ");
    }
    BIASCDOUT(D_CONV_KERNEL, "  sum = "<<sum<<endl);
  }
#endif

  //#define COMPARE
#ifdef COMPARE
  fprintf(stdout, "gauss: \n");
  for (i=-hws; i<=hws; i++){
    fprintf(stdout, "%f ",HVec[i+hws]); 
  }
  fprintf(stdout, "\n");
#endif
}




template <class StorageType, class KernelType>
void Gauss<StorageType, KernelType>::
_AddParam(Param& para, const int grp, const std::string& postfix)
{
  if (para.CheckParam("Sigma"+postfix)){ 
    _GaussSigma=para.GetParamDouble("Sigma"+postfix);
  } else { 
    _GaussSigma=para.AddParamDouble("Sigma"+postfix, 
                                    "standard deviation of gauss curve", 
                                    0.7, 0.0, 10.0, 0, grp);
  }

  if (para.CheckParam("Ratio"+postfix)){ 
    _GaussRatio=para.GetParamDouble("Ratio"+postfix);
  } else { 
    _GaussRatio=para.AddParamDouble("Ratio"+postfix, 
                                    "gauss is truncated, when factor between "
                                    "biggest and smalles entry falls below "
                                    "GaussRatio", 
                                    0.01, 0.0, 1.0, 0, grp);
  }
  para.SetEnhancedFlag("Ratio"+postfix, true);
}
