/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __Gauss_hh__
#define __Gauss_hh__

#include <Utils/Param.hh>

#include "Convolution.hh"

namespace BIAS {
  /** @class Gauss
      @ingroup g_filter
      @brief smoothing with gaussian kernel

      Parameters are sigma and ratio.
      The gaussian kernel is separabel, using vector convolution
      The Gaussian function is truncated, when the ratio between the maximum 
      kernel element and the actual kernel element falls below ratio.

      @author woelk 09/2004 */
  template <class InputStorageType, class OutputStorageType> 
  class Gauss 
    : public Convolution<InputStorageType, OutputStorageType>
  {
  public:
    Gauss();

    Gauss(Param& para, const std::string& postfix="");

    Gauss(Param& para, const int grp, const std::string& postfix);

    ~Gauss();
    
    /** @brief sets gauss kernel if params changed and calls convolution */
    virtual int Filter(const Image<InputStorageType>& src, 
                       Image<OutputStorageType>& dst);

    virtual int Filter3x3(const Image<InputStorageType>& src, 
                          Image<OutputStorageType>& dst);
    virtual int Filter5x5(const Image<InputStorageType>& src, 
                          Image<OutputStorageType>& dst);
  
    /////////////////////////////////////////////////////////////////////////
    
    inline void SetSigma(const double si) { *_GaussSigma=si; }

    inline double GetSigma() const  { return *_GaussSigma; }

    inline void SetRatio(const double ratio) { *_GaussRatio=ratio; }

    inline double GetRatio() const { return *_GaussRatio; } 

    inline void SetHalfWinSize(const int hws)
    { *_GaussSigma=sqrt((double)hws*(double)hws/(-2.0*log(*_GaussRatio))); 
      cerr << "hws "<<hws<<" -> sigma "<<*_GaussSigma<<endl; }


    // these are interfaces for class Tracker is Track is instantiated
    // with out param object
    double *GetSigmaPointer(){ return _GaussSigma;};
    double *GetRatioPointer(){ return _GaussRatio;};

  protected:
    /// the parameter
    double *_GaussSigma;
    double *_GaussRatio;
    // the parameter at the time of the last call of _CalculateKernels 
    // (for caching)
    double _LastSigma, _LastRatio;

    /// calculates the kernel
    void _CalculateKernels(double Sigma, double Ratio);


    void _AddParam(Param& para, const int grp, const std::string& postfix);

  }; // class

} // namespace



#endif // __Gauss_hh__
