/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __CONIC2D_HH__
#define __CONIC2D_HH__
#include "bias_config.h" 

#include <Base/Debug/Error.hh>
#include <Base/Debug/Debug.hh>
#include <Base/Math/Matrix3x3.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Image/Image.hh>
#include <Geometry/Quadric3D.hh>
#include <Geometry/PMatrix.hh>

#define CONIC2D_TYPE double

namespace BIAS {
  /** @class Conic2D
      @ingroup g_geometry
      @brief A 3x3 matrix representing a conic (cone/plane intersection)
      
      A conic is a (contour) line in P2, which resulted from the projection
      of a quadric (see Quadric3D) of P3. Conics are ellipses, hyperbolas, ... 
      Each point x of the conic C fulfills  x^T C x = 0
      which is a quadratic equation in the components of x.
      @author koeser, 10/2003 */

  class Conic2D : public Matrix3x3<CONIC2D_TYPE>
  {

  public:
    /** @brief default constructor sets all fields to zero */
    Conic2D() { SetZero(); }

    /** @brief copy constructor uses Conic2D::operator= */
    Conic2D(const Matrix<CONIC2D_TYPE>& m) { (*this)=m;};

    /** @brief check whether this conic is an ellipse
        ellipses have full rank and thus det!=0 */
    bool IsEllipse();

    /** @brief determines if the point is inside/on/outside the conic

        for "closed" conics, such as ellipses, call this function for
        a point to determine its position relative to the conic:
        <0 means: inside the conic
        >0 means: outside the conic
        =0 means on the border line
        @attention: make sure that the conic is normalized
        LocatePoint internally computes the difference of Mahalanobis distances
        of point2D from the center and a conic point from the center */
    inline double LocatePoint(const HomgPoint2D& point2D) {
      return ( point2D *((*this)*point2D) ); }
      
    /** @brief sets mean of first 2 diag elements to positive value, 
        so that LocatePoint(x)<0 means x is in conic */
    inline bool Normalize() { double d=(*this)[0][0]+(*this)[1][1];  
    if (d==0) return false; (*this) /= d; return true;}

    /** @brief construct an ellipse with explicit parameters
        @param radius_a radius in (originally) x direction
        @param radius_b radius in (originally) y direction
        @param dAngle rotation angle in radiant (math. positive)
        @center center of ellipse  */
    void SetEllipse(const HomgPoint2D& Center, const double& dAngle,
                    const double& radius_a, const double& radius_b);

    /** @brief computes explicit ellipse representation
        
        if the given conic is an ellipse, the ellipse parameters are extracted
        and returned in the functions parameters
        @param dAngle rotation angle in radiant (math. positive)
        @param radius_a radius in (originally) x direction
        @param radius_b radius in (originally) y direction
        @return returned parameters are only valid if true is returned,
                otherwise (*this) is no (or a degenerate) ellipse */
    bool GetEllipseParameters(HomgPoint2D& Center, 
                              double& dAngle,
                              double& radius_a, 
                              double& radius_b);

    /** @brief draws conic into an image using a brute force method */
    void Draw(Image<unsigned char>& img);

    Conic2D& operator=(const Matrix<CONIC2D_TYPE>& m) {
      Matrix3x3<CONIC2D_TYPE>::operator=(m);
      return (*this);
    };

    /** @brief constructs a conic that is a projection of the quadric Q using
        camera matrix P */
    void SetQuadricProjection(const Quadric3D &Q, const PMatrix& P);

    /** @brief returns true if the given Conic intersects with the circle of
        radius Radius around point C, false otherwise

        This is a line intersection ! Does not test whether circle
        lies "in" conic or vice versa ! 
        @note strongly depends upon numerical robustness of SolveQuartic !
        @status tested */
    bool IntersectsCircle(const HomgPoint2D& C, double Radius);
    
    /** @brief returns the dual of this conic, which is the inverse */
    inline Conic2D GetDualConic() const {
      SVD csvd((*this)); return (csvd.Invert()); }

    /** @brief computes signature of matrix, 
        (no. of neg. eigenvalues) - (no. of pos. eigenvalues) or vice versa
        if there are more pos. eigenvalues */
     unsigned int GetSignature();


  protected:
    /** @return a value classifying the conic: 
        -   Discriminant<0: hyperbola 
        -  (Discriminant=0: parabola ?) 
        -   Discriminant>0: ellipse */
    double GetDiscriminant();
  };
  
}

#endif
