/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Base/Image/Image.hh>
#include "Camera.hh"
#include <stdio.h>

#define MAXHEADERLINES 1024

using namespace BIAS;
using namespace std;

template <class StorageType> 
Camera<StorageType>::Camera() : Image<StorageType>()
{
  TimeSec_ = 0;
  TimeUSec_ = 0;
  Orientation_[0] = 0.0;
  Orientation_[1] = 0.0;
  Orientation_[2] = 0.0;
  _f=0.0;
  _C.Set(0.0, 0.0, 0.0);
}

template <class StorageType> 
Camera<StorageType>::~Camera() 
{}

template <class StorageType> 
Camera<StorageType>::Camera(const ImageBase& img) 
  : Image<StorageType>(img) 
{
  TimeSec_ = 0;
  TimeUSec_ = 0;
  Orientation_[0] = 0.0;
  Orientation_[1] = 0.0;
  Orientation_[2] = 0.0;
  _f=0.0;
}

template <class StorageType> 
int Camera<StorageType>::SetPMatrix(double P[3][4])
{
  for (int r = 0; r<3; r++)
    for (int c = 0; c<4; c++)
      P_[r][c] = P[r][c];
  return 0;
}

template <class StorageType>
void Camera<StorageType>::PrintAppData(ostream &os)  
{
  os << "Timestamp   : "<<TimeSec_<<":"<<TimeUSec_<<endl;
  os << "Orientation : "<<Orientation_[0]<<" "<<Orientation_[1]<<" "
     <<Orientation_[2]<<endl;
  os << "P-Matrix: " << endl << P_;
  if (!_R.IsZero()) {
    os << "R-Matrix: " << endl << _R;
  } else {
    if (!P_.IsZero())
      os << "R-Matrix is empty - decomposed from P: " << endl << P_.GetR();
  }
  if (!_K.IsZero()) {
    os << "K-Matrix: " << endl << _K;
  } else {
    if (!P_.IsZero())
      os << "K-Matrix is empty - decomposed from P: " << endl << P_.GetK();
  }
  if (!_C.IsZero()) {
    os << "C-Vector: " << _C << endl;
  } else {
    if (!P_.IsZero())
      os << "C-Vector is empty - decomposed from P: " << P_.GetC() << endl;
  }
  os << "HomgPoint2D: " << _Point << endl;
  os << "Focal length: "<< _f << endl;
#ifdef BIAS_DAIMLERCHRYSLER
  os << ptu << endl << inertial << endl << gps << endl;
#endif
}


template <class StorageType>
int Camera<StorageType>::ParseMetaData()
{
  AppData ad;
  //SetDebugLevel(GetDebugLevel() | D_CAMERA_METADATA);
  
  BIASDOUT(D_CAMERA_METADATA, "_MetaData has " << this->_MetaData.size() 
           << " entries");

  if (this->_MetaData.Find(AppData::MD_PMatrix, "#[PMatrix]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found PMatrix");
    if (ad.tag==AppData::MD_USE_ASCII){
      sscanf(ad.sdata.c_str(),
             "%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
             &P_[0][0], &P_[0][1], &P_[0][2], &P_[0][3], 
             &P_[1][0], &P_[1][1], &P_[1][2], &P_[1][3], 
             &P_[2][0], &P_[2][1], &P_[2][2], &P_[2][3]);
    } else {
      memcpy(P_.GetData(), ad.data, ad.length);
    }
  }
  if (this->_MetaData.Find(AppData::MD_TimeStamp, "#[TimeStamp]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found TimeStamp");
    if (ad.tag==AppData::MD_USE_ASCII){
      sscanf(ad.sdata.c_str(), "%ld %ld " , &TimeSec_, &TimeUSec_);
    } else {
      TimeSec_ = ((long int*)ad.data)[0];
      TimeUSec_ = ((long int*)ad.data)[1];
    }
  }
  if (this->_MetaData.Find(AppData::MD_Orientation, "#[Orientation]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found Orientation");
    if (ad.tag==AppData::MD_USE_ASCII){
      sscanf(ad.sdata.c_str(), "%f %f %f" ,&Orientation_[0], &Orientation_[1], 
             &Orientation_[2]); 
    } else {
      memcpy(&Orientation_, ad.data, ad.length);
    }
  }
  if (this->_MetaData.Find(AppData::MD_CVector, "#[CVector]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found CVector");
    if (ad.tag==AppData::MD_USE_ASCII){
      sscanf(ad.sdata.c_str(), "%lf %lf %lf ", &_C[0], &_C[1], &_C[2]);
    } else {
      memcpy(_C.GetData(), ad.data, ad.length);
    }
  }
  if (this->_MetaData.Find(AppData::MD_RMatrix, "#[RMatrix]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found RMatrix");
    if (ad.tag==AppData::MD_USE_ASCII){
      sscanf(ad.sdata.c_str(), "%lf %lf %lf %lf %lf %lf %lf %lf %lf",
             &_R[0][0], &_R[0][1], &_R[0][2], 
             &_R[1][0], &_R[1][1], &_R[1][2], 
             &_R[2][0], &_R[2][1], &_R[2][2]);
    } else {
      memcpy(_R.GetData(), ad.data, ad.length);
    }
  }
  if (this->_MetaData.Find(AppData::MD_KMatrix, "#[KMatrix]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found KMatrix");
    if (ad.tag==AppData::MD_USE_ASCII){
      sscanf(ad.sdata.c_str(), "%lf %lf %lf %lf %lf %lf %lf %lf %lf",
        &_K[0][0], &_K[0][1], &_K[0][2], 
        &_K[1][0], &_K[1][1], &_K[1][2], 
        &_K[2][0], &_K[2][1], &_K[2][2]);
    } else {
      memcpy(_K.GetData(), ad.data, ad.length);
    }
  }
  if (this->_MetaData.Find(AppData::MD_HomgPoint2D, "#[HomgPoint2D]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found HomgPoint2D");
    if (ad.tag==AppData::MD_USE_ASCII){
      sscanf(ad.sdata.c_str(), "[ %lf %lf %lf ]",
             &_Point[0], &_Point[1], &_Point[2]);
    } else {
      memcpy(_Point.GetData(), ad.data, ad.length);
    }
  }
  if (this->_MetaData.Find(AppData::MD_Focal_Length, "#[Focal Length]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found Focal Length");
    if (ad.tag==AppData::MD_USE_ASCII){
      sscanf(ad.sdata.c_str(), "%lf", &_f);
    } else {
      memcpy(&_f, ad.data, ad.length);
    }
  }
#ifdef BIAS_DAIMLERCHRYSLER
  if (this->_MetaData.Find(AppData::MD_PTU_Data, "#[PTU Data]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found PTU_Data");
    if (ad.tag==AppData::MD_USE_ASCII){
      stringstream ss;
      ss << ad.sdata;
      ss >> ptu;
    } else {
      //memcpy(&ptu, ad.data, ad.length);
      ptu.ReadBinary(ad.data);
    }
    BIASDOUT(D_CAMERA_METADATA, "cam.ptu: "<<ptu<<endl);
  }
  if (this->_MetaData.Find(AppData::MD_Inertial_Sensor,"#[Inertial Sensor]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found Inertial_Sensor");
    if (ad.tag==AppData::MD_USE_ASCII){
      stringstream ss;
      ss << ad.sdata;
      ss >> inertial;
    } else {
      //memcpy(&inertial, ad.data, ad.length);
      inertial.ReadBinary(ad.data);
    }
    BIASDOUT(D_CAMERA_METADATA, "cam.inertiual: "<<inertial<<endl);
  }
  if (this->_MetaData.Find(AppData::MD_Inertial_Sensor,"#[GPS Data]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found GPS Data");
    if (ad.tag==AppData::MD_USE_ASCII){
      stringstream ss;
      ss << ad.sdata;
      ss >> gps;
    } else {
      gps.ReadBinary(ad.data);
    }
    BIASDOUT(D_CAMERA_METADATA, "cam.gps: "<<gps<<endl);
  }
#endif
/*
  if (this->_MetaData.Find(AppData::MD_, "#[]", ad)>=0){
    BIASDOUT(D_CAMERA_METADATA, "found ");
    if (ad.tag==AppData::MD_USE_ASCII){
      sscanf(ad.sdata.c_str(), );
    } else {

    }
  }
*/
  return 0;
}


template <class StorageType>
Camera<StorageType> 
Camera<StorageType>::operator=(const Camera<StorageType> &src) 
{
  this->Image<StorageType>::operator=(src);
  TimeSec_=src.TimeSec_;
  TimeUSec_=src.TimeUSec_;
  Orientation_[0] = src.Orientation_[0];
  Orientation_[1] = src.Orientation_[1];
  Orientation_[2] = src.Orientation_[2];  
  P_=src.P_;  
  _K=src._K;  
  _R=src._R; // rotation of camera
  _C=src._C; // camera center in world coo
  _Point=src._Point;
  _f=src._f;
#ifdef BIAS_DAIMLERCHRYSLER
  ptu=src.ptu;
  inertial=src.inertial;
  gps=src.gps;
#endif
  return (*this);  
}


template <class StorageType>
int Camera<StorageType>::UpdateMetaData()
{
  //AddDebugLevel(D_CAMERA_METADATA);
  char tmpline[256];
  ostringstream os;
  BIASDOUT(D_CAMERA_METADATA,"UpdateMetaData()");
  
  this->_MetaData.Add(AppData::MD_PMatrix, 12*sizeof(double),
                (char*)P_.GetData());
  sprintf(tmpline, "%f %f %f %f %f %f %f %f %f %f %f %f",
          P_[0][0], P_[0][1], P_[0][2], P_[0][3], 
          P_[1][0], P_[1][1], P_[1][2], P_[1][3], 
          P_[2][0], P_[2][1], P_[2][2], P_[2][3]);
  this->_MetaData.Add("#[PMatrix]", tmpline);
  BIASDOUT(D_CAMERA_METADATA,"Added PMatrix");

  long int time[2];
  time[0] = TimeSec_;
  time[1] = TimeUSec_;
  this->_MetaData.Add(AppData::MD_TimeStamp,sizeof(time),(char*)time);
  sprintf(tmpline, "%ld %ld" , TimeSec_, TimeUSec_);
  this->_MetaData.Add("#[TimeStamp]", tmpline);
  BIASDOUT(D_CAMERA_METADATA,"Added ");

  this->_MetaData.Add(AppData::MD_Orientation, sizeof(Orientation_),
                   (char*)Orientation_);
  sprintf(tmpline, "%f %f %f" , Orientation_[0], Orientation_[1], 
          Orientation_[2]);
  this->_MetaData.Add("#[Orientation]", tmpline);
  BIASDOUT(D_CAMERA_METADATA,"Added Orientation");

  this->_MetaData.Add(AppData::MD_CVector, 3*sizeof(double), (char*)_C.GetData());
  sprintf(tmpline, "%f %f %f ", _C[0], _C[1], _C[2]);
  this->_MetaData.Add("#[CVector]", tmpline);
  BIASDOUT(D_CAMERA_METADATA,"Added CVector");

  this->_MetaData.Add(AppData::MD_RMatrix, 9*sizeof(double),
                   (char*)_R.GetData());
  sprintf(tmpline, "%f %f %f %f %f %f %f %f %f",
                 _R[0][0], _R[0][1], _R[0][2], 
                 _R[1][0], _R[1][1], _R[1][2], 
                 _R[2][0], _R[2][1], _R[2][2]);
  this->_MetaData.Add("#[RMatrix]", tmpline);
  BIASDOUT(D_CAMERA_METADATA,"Added RMatrix");

  this->_MetaData.Add(AppData::MD_KMatrix, 9*sizeof(double),
                   (char*)_K.GetData());
  sprintf(tmpline, "%f %f %f %f %f %f %f %f %f",
                 _K[0][0], _K[0][1], _K[0][2], 
                 _K[1][0], _K[1][1], _K[1][2], 
                 _K[2][0], _K[2][1], _K[2][2]);
  this->_MetaData.Add("#[KMatrix]", tmpline);
  BIASDOUT(D_CAMERA_METADATA,"Added KMatrix");

  this->_MetaData.Add(AppData::MD_Focal_Length, sizeof(double),
                (char*)&_f);
  sprintf(tmpline, "%f", _f);
  this->_MetaData.Add("#[Focal Length]", tmpline);
  BIASDOUT(D_CAMERA_METADATA,"Added Focal Length");

#ifdef BIAS_DAIMLERCHRYSLER
  //SetDebugLevel(GetDebugLevel() | D_CAMERA_METADATA);
  int length=ptu.GetBinaryLength();
  char *tmpdata=new char[length];
  ptu.WriteBinary(tmpdata);
  BIASCDOUT(D_CAMERA_METADATA, *(double*)tmpdata<<endl);
  this->_MetaData.Add(AppData::MD_PTU_Data, length, tmpdata);
  BIASDOUT(D_CAMERA_METADATA,"Added PTU Data");
  os.str("");
  os << ptu;
  BIASDOUT(D_CAMERA_METADATA,"ptu "<<os.str());
  this->_MetaData.Add("#[PTU Data]", os.str());
  BIASDOUT(D_CAMERA_METADATA,"Added string PTU Data");

  delete[] tmpdata; tmpdata=NULL;
  length=inertial.GetBinaryLength();
  tmpdata=new char[length];
  inertial.WriteBinary(tmpdata);
  this->_MetaData.Add(AppData::MD_Inertial_Sensor, length, tmpdata);
  os.str("");
  os << inertial;
  this->_MetaData.Add("#[Inertial Sensor]", os.str());
  BIASDOUT(D_CAMERA_METADATA,"Added Inertial Sensor");
  delete[] tmpdata; tmpdata=NULL;

  delete[] tmpdata; tmpdata=NULL;
  length=gps.GetBinaryLength();
  tmpdata=new char[length];
  gps.WriteBinary(tmpdata);
  this->_MetaData.Add(AppData::MD_GPS_Data, length, tmpdata);
  os.str("");
  os << gps;
  this->_MetaData.Add("#[GPS Data]", os.str());
  BIASDOUT(D_CAMERA_METADATA,"Added GPS Data");
  delete[] tmpdata; tmpdata=NULL;
#endif

  BIASDOUT(D_CAMERA_METADATA,"Finished adding meta data");
  return 0;
}
















