/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIASCAMERA_HH__
#define __BIASCAMERA_HH__
#include "bias_config.h" 

#include <Base/Image/Image.hh>
#include <Geometry/PMatrix.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Geometry/RMatrix.hh>
#include <Base/Math/Vector3.hh>
#include <Base/Image/MetaData.hh>

#ifdef BIAS_DAIMLERCHRYSLER
#include "DCData.hh"
#endif

#define D_CAM_DCPTU       0x0001
#define D_CAM_DCINERT     0x0002
#define D_CAMERA_METADATA 0x0004
#define D_CAM_IO          0x0008



namespace BIAS {
/** @class Camera
    @ingroup g_image
    @brief extends the Image by P-Matrix support etc.
    
    This class Camera extends Image with
    - accurate timestamps
    - P-matrix support
    - C, R, K matrix support (camera center, -rotation, -calibration)
    - anything usefull representing a camera
*/
  template <class StorageType> 
  class Camera : public Image<StorageType> 
  {
  public: 
    Camera();
    Camera(const ImageBase& img);
    //    Camera(const Camera<StorageType> &src) { operator=(src);};    
    ~Camera();

    /** After ImageIO::Load() operated on AppData_, this method fills
        P_, Timestamp, DC_*, ... */
    int ParseMetaData();

    /** \brief copy P_ and co. back to MetaData before 
        useful before ImageIO::Write() or ImageIO::ExportImage */
    int UpdateMetaData();
    
    int SetPMatrix(double P[3][4]); 
    int SetPMatrix(const PMatrix matP) {P_=matP; return 0;}; 
    inline int SetP(const PMatrix matP) {P_=matP; return 0;};
    //inline BIAS::PMatrix GetPMatrix() const { return P_; } ; 
    inline BIAS::PMatrix& GetPMatrix() { return P_; } ; 
    inline BIAS::PMatrix& GetP() { return P_; } ;

    inline void SetTime(long int sec, long int usec) 
    { TimeSec_ = sec; TimeUSec_ = usec;}
  
    inline void GetTime(long int &sec, long int &usec) 
    { sec = TimeSec_;usec = TimeUSec_;}
  
    inline void SetOrientation(const float ori[3]) 
    { Orientation_[0] = ori[0];Orientation_[1] = ori[1];
    Orientation_[2] = ori[2];};
  
    inline void SetOrientation(const BIAS::Vector3<double> &ori) { 
      Orientation_[0] = float(ori[0]);Orientation_[1] = float(ori[1]); 
      Orientation_[2] = float(ori[2]); 
    } 
    
    inline void GetOrientation(float &o0,float &o1,float &o2) const {
      o0 = Orientation_[0]; o1 = Orientation_[1]; o2 = Orientation_[2];}

    inline void GetOrientation(BIAS::Vector3<double> &ori) const {
      ori[0] = Orientation_[0]; ori[1] = Orientation_[1]; 
      ori[2] = Orientation_[2];}

    inline void SetC(Vector3<double>& C) { _C = C; };
    //inline Vector3<double> GetC() const { return _C; };
    inline Vector3<double>& GetC() { return _C; };
    
    inline void SetR(RMatrix& R) { _R = R; };
    //inline RMatrix GetR() const { return _R; };
    inline RMatrix& GetR() { return _R; };
    
    inline void SetK(KMatrix& K) { _K = K; };
    //inline KMatrix GetK() const { return _K; };
    inline KMatrix& GetK() { return _K; };
    
    inline void Setf(double focal) {_f = focal; };
    inline double Getf() { return _f; };
    
#ifdef BIAS_DAIMLERCHRYSLER
    inline void SetPTU(DC_ptu& p) { ptu = p; }; 
    inline DC_ptu& GetPTU() { return ptu; };
    
    inline void SetInertial(DC_inertial& i) {inertial = i; }; 
    inline DC_inertial& GetInertial() { return inertial; }; 
    
    inline void SetGPS(DC_GPS& g) {gps = g; }; 
    inline DC_GPS& GetGPS() { return gps; }; 
#endif
    
    Camera<StorageType> operator=(const Camera<StorageType> &src);
    

    void PrintAppData(std::ostream &os) ;

  protected:

    long int TimeSec_;
    long int TimeUSec_;
    float Orientation_[3];
    PMatrix P_;
    KMatrix _K; // camera calibration matrix
    RMatrix _R; // rotation of camera
    Vector3<double> _C; // camera center in world coo
    HomgPoint2D _Point;
    double _f; // focal length in m
#ifdef BIAS_DAIMLERCHRYSLER
    DC_ptu ptu;
    DC_inertial inertial;
    DC_GPS gps;
#endif
    
  };
}
#endif

