/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __INTEGRALHISTOGRAM_HH__
#define __INTEGRALHISTOGRAM_HH__

#include <vector>
#include <Base/Image/Image.hh>
#include <Base/Math/Math.hh>
#include <Base/Math/Vector.hh>
#include "ColorHistogram.hh"


#define D_INT_HIST   (1<<0)

namespace BIAS{

  /**
     @class IntegralHistogram
     This Class handels Integral Histograms on HSL Images
     It represents an integral image histogram  and offers basic methods
     on histograms.
     @author Ingo Schiller 
     @status untested
     @date 25.Oktober 2004
   */

class IntegralHistogram
  {
  public:
    /// 
    IntegralHistogram();
    ~IntegralHistogram();

  
    /**
       Deletes ALL current Histogram information
    */
    void Clear();

    /**
       set the binsize (must be divideable by 2pown)
    */
    void SetBinSize(int binsize);

    /**
       set minsat_, which defines the minimum saturation
    */
    void SetMinSat(int minsat);

    /**
       sets Lambda, which is used to calculate the dist
       in CalcSimilarity
    */
    void SetLambda(int lambda);

    /**
       sets expDist_ true or false, used to calculate the dist
       in CalcSimilarity
    */
    void SetExpDist(bool expdist);
  

    /**
       GetBinSize
    */
    inline int GetBinSize() const {return nBinsize_;} 

    /**
       GetBinSizeSat
    */
    inline int GetBinSizeSat() const {return nBinsizeSat_;}
    
    /**
       Get Histogram
       @param roi = the Region in which the histogram should be created 
       @return = Vector which holds teh Histogram, structure:
       binsizeSat * binsizeHue elements(hue and sat )
        + binsizeSat elements(luminance)
        + totalNumber of entries (last element)
    */
    Vector<int> GetHistogram(int ux,int uy,int lx, int ly);

   
    /** 
        calcluates the histogram from the picture and fills the data structs
        @author Schiller 
        @param image = the image
    */
    int GenerateIntegralHist(BIAS::Image<unsigned char>& image,
                             int nBinsize);


    /** 
        calcluates the histogram from the picture and fills the data structs
        @author Schiller 
        @param image = the image
        @param nBinsizeHue = binsize in hue direction
        @param nBinsizeSat = binsize in sat direction
    */
    int GenerateIntegralHistDiffBin(BIAS::Image<unsigned char>& image,
                             int nBinsizeHue, int nBinsizeSat);
    /**
       Calculates the Bhattacharyya similarity Coeiffient^2
       which is the distance or difference of 2 colorhistograms.   
       @param chist = the colorHistogram we compare with
       @param roi = the roi to compare with: upperx, uppery, lowerx, lowery
       @return double value, which indicates the similarity. 1 if similar.
       @author Schiller oktober 2004
    */
    double CalcSimilarity(ColorHistogram<unsigned char> &chist,
                          int ux,int uy,int lx, int ly);
  
    /**
       dumps information on histogram
       @param os = the outstream, default std::cout
       @author Schiller
    */
    void Dump(std::ostream& os, int y, int x);

  private:
    //this is the major datastruct
    //access via [][], has BIAS::Vector which holds H,S,L in binsize each
    // + bincount as last element in vector
    TNT::Matrix<BIAS::Vector<int> > mHists_;
    std::vector<std::vector<double> > m_cHSData;
    std::vector<double> m_cLData;

    int nMinSat_;
    int nLambda_;
    bool bExpDist_;   
    int nBinsize_;
    int nBinsizeSat_;
    int nOldBinsize_;
    int nOldBinsizeSat_;

    int nShift_; // used to shift the pixel to correct position
    int nShiftSat_;
    inline void CopyNoCheck(const BIAS::Vector<int>& src, 
                            BIAS::Vector<int>& dst){
      //BIASASSERT(src.size() == dst.size());
      register int *srcData = src.GetData(), *dstData = dst.GetData();
      const int *end = srcData + src.size();
      while (srcData<end){
        *dstData++ = *srcData++;
      }
    }
  }; // class


} // namespace MIP
#endif // __INTEGRALHISTOGRAM_HH__
