/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Matcher2D/TrackerSimpleKLT.hh>
#include <Matcher2D/TrackerGaussianWeightedKLT.hh>
#include <Base/Image/Image.hh>
#include <Base/Image/ImageDraw.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Debug/TimeMeasure.hh>
#include <Filter/GradientSobel3x3.hh>
#include <Filter/GradientGauss.hh>
#include <Filter/GradientGaussAsymmetric.hh>
#include <Filter/Gauss.hh>

#include <iostream>
#include <iomanip>

using namespace BIAS;
using namespace std;

// ./ExampleTrackerBase 157 157 ~/cvs/TestData/tracker_noise/image*.pgm

#define type 0
#if type == 0
#define StorageType float
#endif

#define DEFAULT_PARA_FILE_NAME "TrackerBase.ini"

int main(int argc, char*argv[])
{
  Param para;
  Image<StorageType> fim[2], im[2], drawim;
  Image<StorageType> gradx[2], grady[2];
  KLT_TYPE error, err, residuum, errorsum=0.0;
  HomgPoint2D p[2];
  unsigned int uip[2], rectsize=7;
  int newim, oldim, iter;
  int res, counter;
  bool debug=true;
  string file;
  bool draw=false;
  MetaData *md;
  AppData appdata;
  HomgPoint2D realpoint;

  TrackerBase<StorageType> *tracker=NULL;
  TrackerSimpleKLT<StorageType> ts(para);
  TrackerGaussianWeightedKLT<StorageType> tgw(para);
  int *trackertype=TrackerBase<StorageType>::AddTrackerType(para);
  FilterNTo2N<StorageType, StorageType> *grad=NULL;
  GradientSobel3x3<StorageType, StorageType> gsobel3x3(para);
  GradientGauss<StorageType,StorageType> ggauss(para);
  GradientGaussAsymmetric<StorageType, StorageType> 
    ggaussasymmetric(para);
  int *gradienttype = AddGradientType(para);
  Gauss<StorageType, StorageType> gauss(para);

  TimeMeasure timer, tc;

  int argind = para.UpdateParameter(argc, argv, DEFAULT_PARA_FILE_NAME);

  if (argc-argind<4 || argind<1){
    cerr << argv[0] << " : <x-coo> <y-coo> <image> <image> ..."<< endl;
    return -5;
  }

  switch (*trackertype){
  case TT_SimpleKLT:
    tracker = &ts;
    break;
  case TT_GaussianWeightedKLT:
    tracker = &tgw;
    break;
  default:
    BIASERR("invalid tracker type");
    tracker = NULL;
    break;
  }

  //tracker->AddDebugLevel(D_TRACKERB_INIT);
  //tracker->AddDebugLevel(D_TRACKERB_SKLT);
  //tracker->AddDebugLevel(D_TRACKERB_GKLT);
  //tracker->AddDebugLevel(D_TRACKERB_BILINEAR);

  switch (*gradienttype){
  case GT_Sobel3x3:
    grad = &gsobel3x3;
    cerr <<"Sobel3x3\n";
    break;
  case GT_Gauss:
    grad = &ggauss;
    cerr <<"GradientGauss\n";
    break;
  case GT_GaussAsymmetric:
    grad = &ggaussasymmetric;
    cerr <<"GradientGaussAsymmetric\n";
    break;
  default:
    BIASERR("unknown gradient type");
    grad = NULL;
    break;
  }
  grad->AddDebugLevel(D_GRADGAUSS_KERNEL);
  grad->AddDebugLevel(D_CONV_KERNEL);
  
  p[0][2]=p[1][2]=1.0;
  p[0][0]=atof(argv[argind]);
  p[0][1]=atof(argv[argind+1]);
  cerr << p[0] << endl;

#if type == 0
  if (ImageIO::Load(argv[argind+2], im[0])!=0){
#else
  if (ImageIO::Load(argv[argind+2], ucim)!=0){
#endif
    BIASERR("error loading image "<<argv[argind+2]);
    return -1;
  }
#if type != 0
  if (ImageConvert::ConvertST(ucim, im[0], ImageBase::ST_float)!=0){
      BIASERR("error converting image "<<argv[i]);
  }
  ImageIO::Write("fimage.mip", im[0]);
#endif

  grad->Filter(im[0], gradx[0], grady[0]);
  gauss.Filter(im[0], fim[0]);

  counter=argind+2;
  newim=0;
  cerr << "argind: "<<argind<<"\targc: "<<argc<<endl;
  while (counter < argc-1){
    counter++;
    newim=(newim==0)?(1):(0);
    oldim=(newim==0)?(1):(0);
    
#if type == 0
    if (ImageIO::Load(argv[counter], im[newim])!=0){
#else
    if (ImageIO::Load(argv[counter], ucim)!=0){
#endif
      BIASERR("error loading image "<<argv[counter]);
      return -1;
    } else {
      if (debug) 
        cout << "--------------- newim: " << argv[counter] 
             << "  --- oldim: " << argv[counter-1] << endl;
      
      md=im[newim].GetMetaData();
      if (md->Find(AppData::MD_HomgPoint2D, "#[HomgPoint2D]", appdata)>=0){
        if (appdata.tag==AppData::MD_USE_ASCII){
          if (debug) cerr << appdata.stag <<" : "<<appdata.sdata<<endl;
          sscanf(appdata.sdata.c_str(), "[ %lf %lf %lf ]",
                 &realpoint[0], &realpoint[1], &realpoint[2]);
        } else {
          memcpy(realpoint.GetData(), appdata.data, appdata.length);
        }
        if (debug) cerr <<"realpoint is at "<<realpoint<<endl;
      } else {
        BIASERR("cannot find HomgPoint2D in meta data");
      }
    }
#if type != 0
    if (ImageConvert::ConvertST(ucim, im[newim], ImageBase::ST_float)!=0){
      BIASERR("error converting image "<<argv[i]);
    }
    ImageIO::Write("fimage.mip", im[newim]);
#endif
    
    grad->Filter(im[newim], gradx[newim], grady[newim]);
    gauss.Filter(im[newim], fim[newim]);
    
    timer.Start();
    tc.Start();

#ifdef BIAS_DEBUG    
    ImageIO::Write("im1.mip", im[oldim]);
    ImageIO::Write("im2.mip", im[newim]);
    ImageIO::Write("fim1.mip", fim[oldim]);
    ImageIO::Write("fim2.mip", fim[newim]);
    ImageIO::Write("gx1.mip", gradx[oldim]);
    ImageIO::Write("gx2.mip", gradx[newim]);
    ImageIO::Write("gy1.mip", grady[oldim]);
    ImageIO::Write("gy2.mip", grady[newim]);
    cerr << "Tracking from "<<p[oldim]<<"  -->  "<<p[oldim]<<endl;
#endif  
    tracker->Init(fim[oldim], fim[newim], gradx[oldim], grady[oldim], 
                  gradx[newim], grady[newim]);

    res = tracker->Track(p[oldim], p[oldim], p[newim], error, iter, residuum);
    
    tc.Stop();
    timer.Stop();

    tc.Print();
    tc.Reset();

    if (res<0){
      BIASERR("error matching "<<res);
      if (res==-2){
        cout << "point ("<<p[oldim][0]<<", "<<p[oldim][1]<<") too close to "
           <<"image border, stopping"<<endl;
        break;
      }
      assert(false);
      
    }
    err=sqrt((p[newim][0]-realpoint[0])*(p[newim][0]-realpoint[0])+
             (p[newim][1]-realpoint[1])*(p[newim][1]-realpoint[1]));
    errorsum+=err;

    if (draw){
      uip[0]=(unsigned int)rint(p[newim][0]);
      uip[1]=(unsigned int)rint(p[newim][1]);
      drawim=im[newim];
      ImageDraw<StorageType>::RectangleCenter(drawim, uip, rectsize);
 
      file=argv[counter];
      file+=".tracked";
      ImageIO::Write(file, drawim);
    }
    cout << setw(2) << counter-3 << " : (" << p[oldim][0]<<", "<<p[oldim][1]
         << ") -> ("<< p[newim][0]<<", "<<p[newim][1]<<")  klt-error: "<<error
         <<"   real: ("<<realpoint[0]<<", "<<realpoint[1]
         <<")   dist-err: "<<err<<"   residuum: "<<residuum<<endl;  

  }
  cerr << "errorsum over "<<counter-3<<" images: "<<errorsum<<endl
       << "  mean error "<<errorsum/(double)(counter-3)<<endl;

  timer.Print();
  cerr << "KLT trackin over "<<argc-3<<" images took "<<timer.GetRealTime()
       <<" us"
       << "\n  this is equal to "<<timer.GetRealTime()/double(argc-3)
       <<" us per image and corner" << endl;
  return 0;
}

