/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel

BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __TrackerSimpleKLT__hh__
#define __TrackerSimpleKLT__hh__

#include "TrackerBase.hh"
#include <Base/Math/Matrix2x2.hh>
#include <Base/Math/Matrix.hh>
#include <Base/Math/Vector2.hh>

namespace BIAS {


  /** @class TrackerSimpleKLT
      @ingroup g_geometry
      @brief the gradient decent tracker as developed initially by 
      Lucas and Tomasi and described in Birchfeld "Derivation of
      Kanade-Lucas-Tomasi Tracking Equation" 

      @author woelk 09/2004 */
  template <class StorageType>
  class DLLEXPORTAPI TrackerSimpleKLT : public TrackerBase<StorageType>
  {
  public :
    TrackerSimpleKLT();

    TrackerSimpleKLT(Param& para, const std::string& postfix="");

    TrackerSimpleKLT(Param& para, const int grp, const std::string& postfix);

    ~TrackerSimpleKLT();

    /** see TrackerBase::Init() for description 
        @author woelk 09/2004  */
    void Init(Image<StorageType>& im1, Image<StorageType>& im2, 
              Image<StorageType>& gradx1, Image<StorageType>& grady1, 
              Image<StorageType>& gradx2, Image<StorageType>& grady2);

    /** see TrackerBase::Track() for description 
        @returns  0 on success (error< maxerror)
                 -1 when no spatial structure is present
                 -2 when point slid out of image
                 -3 when maxiter is reached and error is above maxerror
                 -4 when a point lies outside of valid region in images
                 -5 when the mean absolute grey difference (resiuum) between 
                      the two points is bigger than *_MaxResiduum
                 -6 when p1[i] is at infinity 
        @author woelk 09/2004  */
    virtual void Track(std::vector<HomgPoint2D>& p1, 
                       std::vector<HomgPoint2D>& p2, 
                       std::vector<HomgPoint2D>& p2tracked, 
                       std::vector<KLT_TYPE>& error, 
                       std::vector<int>& numiter,
                       std::vector<KLT_TYPE>& residui, 
                       std::vector<int>& res);

    /** see TrackerBase::Track() for description 
        @author woelk 09/2004  */
    virtual int Track(HomgPoint2D& p1, HomgPoint2D& p2, HomgPoint2D& result, 
                      KLT_TYPE& error, int &iter, KLT_TYPE& resiuduum);

    //////////////////////////////////////////////////////////////////////
    // basic functions, needed when images are no longer present
    //////////////////////////////////////////////////////////////////////

    /** see TrackerBase::Track() for description 
        @author woelk 09/2004 */
    virtual void Init(StorageType **ida1, StorageType **ida2, 
                      StorageType **gradx1, StorageType **grady1, 
                      StorageType **gradx2, StorageType **grady2, 
                      unsigned int width, unsigned int height,
                      unsigned gradient_half_mask);

    /** see TrackerBase::Track() for description 
        @returns  0 on success (error< maxerror)
                 -1 when no spatial structure is present
                 -2 when point slid out of image
                 -3 when maxiter is reached and error is above maxerror
                 -4 when a point lies outside of valid region in images
                 -5 when the mean absolute grey difference (resiuum) between 
                      the two points is bigger than *_MaxResiduum
        @author woelk 09/2004 */
    virtual int Track(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], 
                      KLT_TYPE& error, int &iter, KLT_TYPE& resiuduum);

  protected:
    Matrix2x2<KLT_TYPE> _G, _Ginv;
    Vector2<KLT_TYPE> _e, _d, _disp;
    Matrix<KLT_TYPE> _gx, _gy, _bl1, _bl2, _gx1, _gy1, _gx2, _gy2, _gsx, _gsy;
    int _LastHalfWinSize, _WinSize;

    /** the actual computation for a single point is done in this function
        @author woelk 09/2004 */
    int Track_(KLT_TYPE p1[2], KLT_TYPE p2[2], KLT_TYPE result[2], 
               KLT_TYPE& error, int &iter, KLT_TYPE& resiuduum);

    inline void _Resize(const int halfwinsize){
      _LastHalfWinSize=halfwinsize;
      _WinSize=(halfwinsize<<1)+1;
      _gx1.newsize(_WinSize, _WinSize);
      _gy1.newsize(_WinSize, _WinSize);
      _gx2.newsize(_WinSize, _WinSize);
      _gy2.newsize(_WinSize, _WinSize);
      _gsx.newsize(_WinSize, _WinSize);
      _gsy.newsize(_WinSize, _WinSize);
      _bl1.newsize(_WinSize, _WinSize);
      _bl2.newsize(_WinSize, _WinSize);
      ;
    }

    inline void _MyBilinearRegion1(KLT_TYPE x, KLT_TYPE y, int hws);

    inline void _MyBilinearRegion2(KLT_TYPE x, KLT_TYPE y, int hws);

    // static variables needed in _MyBilienarRegion
    unsigned rf, rc, cf, cc, r, c, x_floor, y_floor, x_floor_end, y_floor_end;
    KLT_TYPE ul, ur, ll, lr, brdx, brdy, brdxy;

  }; // class


  template <class StorageType>
  inline void TrackerSimpleKLT<StorageType>::
  _MyBilinearRegion1(KLT_TYPE x, KLT_TYPE y, int hws)
  {
    x_floor = (unsigned int)floor(x);
    y_floor = (unsigned int)floor(y);
    x_floor_end = x_floor+hws+1;
    y_floor_end = y_floor+hws+1;
    brdy = y - (KLT_TYPE)(y_floor);
    brdx = x - (KLT_TYPE)(x_floor);
    brdxy = brdx*brdy;
    for (rf=y_floor-hws, rc=rf+1, r=0; rf<y_floor_end; rf++, rc++, r++){
      for (cf=x_floor-hws, cc=cf+1, c=0; cf<x_floor_end; cf++, cc++, c++){
        ul = (KLT_TYPE)this->_ida1[rf][cf];
        ur = (KLT_TYPE)this->_ida1[rf][cc];
        ll = (KLT_TYPE)this->_ida1[rc][cf];
        lr = (KLT_TYPE)this->_ida1[rc][cc];
        _bl1[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_gradx1[rf][cf]);
        ur = (KLT_TYPE)(this->_gradx1[rf][cc]);
        ll = (KLT_TYPE)(this->_gradx1[rc][cf]);
        lr = (KLT_TYPE)(this->_gradx1[rc][cc]);
        _gx1[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_grady1[rf][cf]);
        ur = (KLT_TYPE)(this->_grady1[rf][cc]);
        ll = (KLT_TYPE)(this->_grady1[rc][cf]);
        lr = (KLT_TYPE)(this->_grady1[rc][cc]);
        _gy1[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
      }
    }
  }

  template <class StorageType>
  inline void TrackerSimpleKLT<StorageType>::
  _MyBilinearRegion2(KLT_TYPE x, KLT_TYPE y, int hws)
  {
    x_floor = (unsigned int)floor(x);
    y_floor = (unsigned int)floor(y);
    x_floor_end = x_floor+hws+1;
    y_floor_end = y_floor+hws+1;
    brdy = y - (KLT_TYPE)(y_floor);
    brdx = x - (KLT_TYPE)(x_floor);
    brdxy = brdx*brdy;
    for (rf=y_floor-hws, rc=rf+1, r=0; rf<y_floor_end; rf++, rc++, r++){
      for (cf=x_floor-hws, cc=cf+1, c=0; cf<x_floor_end; cf++, cc++, c++){
        ul = (KLT_TYPE)this->_ida2[rf][cf];
        ur = (KLT_TYPE)this->_ida2[rf][cc];
        ll = (KLT_TYPE)this->_ida2[rc][cf];
        lr = (KLT_TYPE)this->_ida2[rc][cc];
        _bl2[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_gradx2[rf][cf]);
        ur = (KLT_TYPE)(this->_gradx2[rf][cc]);
        ll = (KLT_TYPE)(this->_gradx2[rc][cf]);
        lr = (KLT_TYPE)(this->_gradx2[rc][cc]);
        _gx2[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(this->_grady2[rf][cf]);
        ur = (KLT_TYPE)(this->_grady2[rf][cc]);
        ll = (KLT_TYPE)(this->_grady2[rc][cf]);
        lr = (KLT_TYPE)(this->_grady2[rc][cc]);
        _gy2[r][c]=(ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
      }
    }
  }

} // namespace


#endif // __TrackerSimpleKLT__hh__
