/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "FFT.hh"

#include <Base/Debug/Error.hh>

#include <complex>
#include <iomanip>

using namespace BIAS;
using namespace std;

FFT::FFT()
{
  _Size=0;
  _in=NULL;
  _out=NULL;
  _p=NULL;
}

FFT::~FFT()
{
  Release();
}

void FFT::Init(int size)
{
#ifdef BIAS_DEBUG
  if (_Size!=0 || _InputSize!=0 || _OutNum!=0 || _out || _in || _p){
    BIASERR("FFT: call Release() before calling Init() a second time");
    BIASASSERT(false);
  }
#endif
  _Size=size;
  _InputSize=_Size*sizeof(double);
  _OutNum=(_Size>>1)+1;
  _out =(fftw_complex*)fftw_malloc(sizeof(fftw_complex) * _OutNum);
  _in = (double*)fftw_malloc(sizeof(double) * _Size);
  // this is slow a few seconds
  _p = fftw_plan_dft_r2c_1d(_Size, _in, _out, FFTW_MEASURE);
}

void FFT::Release()
{
  if (_p) fftw_destroy_plan(_p);
  if (_in) fftw_free(_in); 
  if (_out) fftw_free(_out);
  _in=NULL;
  _out=NULL;
  _p=NULL;
  _Size=_InputSize=_OutNum=0;
}

void FFT::Compute(double *in)
{
  memcpy((void *)_in, (void *)in, _InputSize); 
  fftw_execute(_p);
}

void FFT::Compute(Vector<double> &in)
{
#ifdef BIAS_DEBUG
  if ((int)in.Size()!=_Size){
    BIASERR("called FFT::Compute() with vector of wrong size "<<in.Size());
    BIASASSERT(false);
  }
#endif
  Compute(in.GetData());
}

void FFT::GetPhase(double *phase)
{
  complex<double> *p=reinterpret_cast<complex<double> *>(_out), *e=p+_OutNum;
  double *m=phase;
  while (p<e) {
    *m=arg(*p);
    m++;
    p++;
  }
}

void FFT::GetMagnitude(double *mag)
{
  complex<double> *p=reinterpret_cast<complex<double> *>(_out), *e=p+_OutNum;
  double *m=mag;
  while (p<e) {
    *m=abs(*p);
    m++;
    p++;
  }
}

void FFT::GetPhase(Vector<double>& phase)
{
  if ((int)phase.Size()!=_OutNum){
    phase.newsize(_OutNum);
  }
  complex<double> *p=reinterpret_cast<complex<double> *>(_out);
  for (int i=0; i<_OutNum; i++){
    phase[i]=arg(p[i]);
  }
}

void FFT::GetMagnitude(Vector<double>& mag)
{
  if ((int)mag.Size()!=_OutNum){
    mag.newsize(_OutNum);
  }
  complex<double> *p=reinterpret_cast<complex<double> *>(_out);
  for (int i=0; i<_OutNum; i++){
    mag[i]=abs(p[i]);
  }
}

void FFT::DumpResult(std::ostream& os)
{
  Vector<double> mag, phase;
  GetMagnitude(mag);
  GetPhase(phase);
  os << "#result of fft: "<<endl;
  os <<setw(4)<<"#num"<<setw(15)<<"freq"<<setw(15)<<"mag"<<setw(15)
       <<"phase"<<endl;
  for (int i=0; i<_OutNum; i++){
    os <<setw(4)<<i
         <<setw(15)<<(double)i/(double)_Size
         <<setw(15)<<mag[i]
         <<setw(15)<<phase[i]<<"\n";
  }

//   os << "# pipe outputt into file: my_program_with_FFT > tmp.dat"<<endl
//      << "# and view with gnuplot : echo \"plot \\\"tmp.dat\\\" u 2:3 t \\\"magnitude\\\" w l, \\\"tmp.dat\\\" u 2:4 t \\\"phase\\\" w l; pause 99999;\" |gnuplot"<<endl;

}
