#ifndef __MSAC_hh__
#define __MSAC_hh__

#include <Base/Debug/Debug.hh>
#include <Base/Math/Random.hh>
#include <Utils/Param.hh>

#include <vector>
#include <iterator>

namespace BIAS {

#define D_MSAC_SOLVEM  0x00000001
#define D_MSAC_SAMPLES 0x00000002
#define D_MSAC_INIT    0x00000004

  template <class SolutionType>
  class MSAC : public virtual BIAS::Debug
  {
  public:
    MSAC(BIAS::Param& para);
    MSAC(BIAS::Param& para, const int grp, const std::string& postfix);

    virtual ~MSAC();

    int Init(int data_size);

    int SolveMaster(SolutionType &solution,
                    std::vector<bool> &inliers);

    virtual 
    int GetSampleSolution(std::vector<unsigned int> &which_samples,
                          SolutionType &solutions);

    virtual 
    bool EvaluateSolution(const SolutionType& solution,
                          std::vector<bool>& inlier, int& num_inliers,
                          double& evaluate_score);
    
    virtual 
    bool RefineSolution(std::vector<unsigned int>& which_samples,
                        SolutionType& solution,std::vector<bool>& new_inliers);

    virtual 
    bool GenerateSamples(int sample_index, 
                         std::vector<unsigned int>& which_samples);
    
 
  protected:
    bool *_MSACRefineSolution;
    double *_MSACConfidenceThresh; 
    double *_MSACSigmaFac; 
    double *_dExpectedSigma;
    double *_dExpectedInlierFraction;
    int _iDataSize; // number of data
    int _iSolutionCount; // number of solutions to evaluate
    int *_iSampleSize; // size of minimum data set needed to calc. sol
    double _dTerminateScore;
    
    bool _GenerateSamplesRandom(int sample_index,
                                std::vector<unsigned int>& which_samples);

    double _CalculateTerminateScore(double sigma, 
                                    double expected_inlier_fraction);

    int _CalculateSampleCount(double expected_inlier_fraction);


    void _AddParameter(BIAS::Param& para, int grp, 
                       const std::string& postfix);

   }; // class





  //////////////////////////////////////////////////////////////////////////
  //       implementation                    
  //////////////////////////////////////////////////////////////////////////

  template <class SolutionType>
  MSAC<SolutionType>::MSAC(BIAS::Param& para)
    : BIAS::Debug()
  {
    _MSACRefineSolution=NULL;
    _MSACConfidenceThresh=_MSACSigmaFac=NULL;
    _iDataSize=0;
    _iSolutionCount=0;
    _iSampleSize=NULL;
    _dTerminateScore=0.0;
    _dExpectedSigma=NULL;
    _dExpectedInlierFraction=NULL;
    int grp=para.GetFreeGroupID();
    cerr << "using id "<<grp<<" for MSAC\n";

    _AddParameter(para, grp, "-MSAC");
  }

  template <class SolutionType>
  MSAC<SolutionType>::MSAC(BIAS::Param& para, int grp, 
                           const std::string& postfix)
    : BIAS::Debug()
  {
    _MSACRefineSolution=NULL;
    _MSACConfidenceThresh=_MSACSigmaFac=NULL;
    _iDataSize=0;
    _iSolutionCount=0;
    _iSampleSize=NULL;
    _dTerminateScore=0.0;
    _dExpectedSigma=NULL;
    _dExpectedInlierFraction=NULL;
    _AddParameter(para, grp, "-MSAC"+postfix);
  }


  template <class SolutionType>
  MSAC<SolutionType>::~MSAC()
  {}

  
  template <class SolutionType>
  int MSAC<SolutionType>::Init(int data_size)
  {
    _iDataSize=data_size;
    _dTerminateScore=_CalculateTerminateScore(*_dExpectedSigma, 
                                              *_dExpectedInlierFraction);
    //    BIASERR("fast termination disabled"); _dTerminateScore=0.0;
    _iSolutionCount=_CalculateSampleCount(*_dExpectedInlierFraction);
    BIASCDOUT(D_MSAC_INIT, _iDataSize<<" data points whereof "
              <<*_dExpectedInlierFraction*data_size<<" are expected to be "
              <<"inliers.\n Inliers are assumed normal distributed with sigma "
              <<*_dExpectedSigma<<".\n Using confidence in solution of "
              <<*_MSACConfidenceThresh
              <<" % with sample set size "<<*_iSampleSize<<" leads to at most "
              <<_iSolutionCount
              <<" solutions to be evaluated.\n MSAC will abort on score below "
              <<_dTerminateScore<<"\n");
    BIASASSERT(_iDataSize>=*_iSampleSize);
    return 0;
  }


  template <class SolutionType>
  int MSAC<SolutionType>::SolveMaster(SolutionType& solution,
                                      std::vector<bool>& inliers)
  {
    SolutionType best_sol, act_sol;
    std::vector<unsigned> which_samples;
    std::vector<bool> tmp(_iDataSize, false);
    std::vector<std::vector<bool> > inliervec(_iSolutionCount, tmp);
    double act_score, best_score=DBL_MAX;
    int sample_index, best_index=-1, act_num_inlier, best_num_inlier=0;

    which_samples.resize(*_iSampleSize);
    inliers.resize(_iSolutionCount);

    for (sample_index = 0; sample_index < _iSolutionCount; sample_index++){
      //for (sample_index = 0; sample_index < 1; sample_index++){
      if (GenerateSamples(sample_index, which_samples)){
        if (!GetSampleSolution(which_samples, act_sol)){
          BIASERR("GetSampleSolution failed");
          BIASASSERT(false);
        } 
        if (!EvaluateSolution(act_sol, inliervec[sample_index], 
                              act_num_inlier, act_score)){
          BIASERR("Evaluate Solution failed");
          BIASASSERT(false);
        }
        BIASCDOUT(D_MSAC_SOLVEM, setw(3)<<sample_index<<"th solution: "
                  <<act_sol<<" with score "<<act_score<<" and "<<act_num_inlier
                  <<" inliers\n");

        if (*_MSACRefineSolution && true){
          which_samples.clear();
          for (int i=0; i<_iDataSize; i++){
            if (inliervec[sample_index][i]){
              which_samples.push_back(i);
            }
          }
          
          if (!RefineSolution(which_samples, act_sol, 
                              inliervec[sample_index])){
            BIASERR("RefineSolution returned false");
            BIASASSERT(false);
          } 
          if (!EvaluateSolution(act_sol, inliervec[sample_index], 
                                act_num_inlier, act_score)){
            BIASERR("Evaluate Solution failed");
            BIASASSERT(false);
          }
          BIASCDOUT(D_MSAC_SOLVEM, setw(3)<<sample_index
                    <<"th refined solution: "
                    <<act_sol<<" with score "<<act_score<<" and "
                    <<act_num_inlier<<" inliers\n");
        }
        
        
#ifdef BIAS_DEBUG
        if (sample_index!=0){
          BIASCDOUT(D_MSAC_SOLVEM, "(best: "<<best_sol<<"/"<<best_score<<"/"
                    <<best_num_inlier<<endl);
        }
#endif
        if (best_score>act_score){
          best_score=act_score;
          best_index=sample_index;
          best_num_inlier=act_num_inlier;
          best_sol=act_sol;
        }
        if (best_score<_dTerminateScore){
          BIASCDOUT(D_MSAC_SOLVEM, "finished sampling, got a good score "
                    <<best_score<<endl);
          break;
        }
      } else { // if (GenerateSamples(sample_index, which_samples)){
        BIASERR("GenerateSamples failed");
        BIASASSERT(false);
      }
    }
    // now search best solution and refine it if not already done in loop
    if (!*_MSACRefineSolution){
      solution=best_sol;
      BIASCDOUT(D_MSAC_SOLVEM, "now refining solution "<<solution<<endl);
      which_samples.clear();
      for (int i=0; i<_iDataSize; i++){
        if (inliervec[best_index][i]){
          which_samples.push_back(i);
        }
      }
      
      if (!RefineSolution(which_samples, solution, inliers)){
        BIASERR("RefineSolution returned false");
        BIASASSERT(false);
      } 
    } else {
      inliers=inliervec[best_index];
      solution=best_sol;
    }    
    BIASCDOUT(D_MSAC_SOLVEM, "finished: score "<<best_score<<" with "
              <<best_num_inlier<<" inlier, solution is "<<solution<<endl);
    return 0;
  }
  
  template <class SolutionType>
  int MSAC<SolutionType>::
  GetSampleSolution(std::vector<unsigned int> &which_samples, 
                    SolutionType &solutions)
  {
    BIASERR("this function should be implemented in derived class");
    return -1;
  }



  template <class SolutionType>
  bool MSAC<SolutionType>::
  EvaluateSolution(const SolutionType& solution, std::vector<bool>& inlier, 
                   int& num_inliers, double& evaluate_score)
  {
    BIASERR("this function should be implemented in derived class");
    return false;
  }




  template <class SolutionType>
  bool MSAC<SolutionType>::
  RefineSolution(std::vector<unsigned int>& which_samples, 
                 SolutionType& solution, std::vector<bool>& new_inliers)
  {
    BIASERR("this function should be implemented in derived class");
    return false;
  }



  template <class SolutionType>
  bool MSAC<SolutionType>::
  GenerateSamples(int sample_index, std::vector<unsigned int>& which_samples)
  {
    return _GenerateSamplesRandom(sample_index, which_samples);
  }
  



  template <class SolutionType>
  bool MSAC<SolutionType>::
  _GenerateSamplesRandom(int sample_index, 
                         std::vector<unsigned int>& which_samples)
  {
    bool res=true;
#ifdef BIAS_DEBUG
    if (*_iSampleSize==0){
      BIASERR("call Init or appropriate constructor befor using MSAC");
      BIASASSERT(false);
      res=false;
    }
    if (*_iSampleSize>=_iDataSize){
      BIASERR("_iDataSize should be set in derived class befor using MSAC"
              <<" _iSampleSize="<<*_iSampleSize<<"  _iDataSize : "
              <<_iDataSize);
      BIASASSERT(false);
      res=false;
    }
#endif
    // initialzing random with 0 and 1 does not seem to work well
    BIAS::Random ran(sample_index+rand());
    int count=0;
    unsigned int index;
    which_samples.clear();

    do {
      index = ran.GetUniformDistributedInt(0, _iDataSize-1);
      if (find(which_samples.begin(), which_samples.end(), index)==
          which_samples.end()){
        which_samples.push_back(index);
        count++;
      }
    } while (count<*_iSampleSize);

#ifdef BIAS_DEBUG
    if ((GetDebugLevel() & D_MSAC_SAMPLES)!=0){
      BIASCDOUT(D_MSAC_SAMPLES, "---------- sample "<<sample_index
               <<" ---------------------------------"<<"\nnow using data: ");
      std::ostream_iterator<int> oi(GetDebugStream(), " ");
      std::copy(which_samples.begin(), which_samples.end(), oi);
      GetDebugStream() << std::endl;
    }
#endif

    return res;
  }


  
  
  template <class SolutionType>
  double MSAC<SolutionType>::
  _CalculateTerminateScore(double sigma, double expected_inlier_fraction)
  {
    double res=0.0;
    res += _iDataSize * expected_inlier_fraction * sigma;
    res += _iDataSize * (1.0-expected_inlier_fraction) *  
      *_MSACSigmaFac * sigma;
    return res;
  }




  template <class SolutionType>
  int MSAC<SolutionType>::
  _CalculateSampleCount(double expected_inlier_fraction)
  {
    if (expected_inlier_fraction == 1.0)
      // the analytic expression below cant cope with this so do it explicitly.
      return 1;

    // from Torr's thesis.
    return int(log (1.0 - (*_MSACConfidenceThresh)) /
               log (1.0 - pow (expected_inlier_fraction, 
                                     (double) *_iSampleSize)));

  }







  template <class SolutionType>
  void MSAC<SolutionType>::
  _AddParameter(BIAS::Param& para, int grp, const std::string& postfix)
  {

    if (para.CheckParam("RefineSolution"+postfix)){ 
      _MSACRefineSolution=para.GetParamBool("RefineSolution"+postfix);
    } else {
      _MSACRefineSolution=
        para.AddParamBool("RefineSolution"+postfix, 
                          "should the solution be refined afterwards using all inliers?", 
                          false, 0, grp); 
    } 

    if (para.CheckParam("ConfidenceThresh"+postfix)){ 
      _MSACConfidenceThresh=para.GetParamDouble("ConfidenceThresh"+postfix);
    } else {
      _MSACConfidenceThresh=
        para.AddParamDouble("ConfidenceThresh"+postfix, 
                            "confidence in a correct solution", 
                            0.95, 0.0, 1.0, 0, grp); 
    }

    if (para.CheckParam("SigmaFac"+postfix)){ 
      _MSACSigmaFac=para.GetParamDouble("SigmaFac"+postfix);
    } else {
      _MSACSigmaFac=
        para.AddParamDouble("SigmaFac"+postfix, 
                            "a datum is considered inlying, when the residuum is below SigmaFac*sigma", 
                            1.96, 0.0, DBL_MAX, 0, grp); 
    }

    if (para.CheckParam("ExpectedSigma"+postfix)){ 
      _dExpectedSigma=para.GetParamDouble("ExpectedSigma"+postfix);
    } else {
      _dExpectedSigma=
        para.AddParamDouble("ExpectedSigma"+postfix, 
                            "expected noise of inlying data", 0.0002, 
                            0.0, DBL_MAX, 0, grp); 
    }

    if (para.CheckParam("ExpectedInlierFraction"+postfix)){ 
      _dExpectedInlierFraction=
        para.GetParamDouble("ExpectedInlierFraction"+postfix);
    } else {
      _dExpectedInlierFraction=
        para.AddParamDouble("ExpectedInlierFraction"+postfix, 
                            "expected inlier percentage", 
                            0.8, 0.0, 1.0, 0, grp); 
    }

    if (para.CheckParam("SampleCount"+postfix)){ 
      _iSampleSize=
        para.GetParamInt("SampleCount"+postfix);
    } else {
      _iSampleSize=
        para.AddParamInt("SampleCount"+postfix, 
                         "required number of samples for calulation of solution", 
                         1, 1, INT_MAX, 0, grp); 
    }

     

  }



} // namespace

#endif // __MSAC_hh__
