/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "SVD.hh"

#include <algorithm>
#include <cmath>

#include "Lapack.hh"

using namespace BIAS;
using namespace std;

SVD::SVD(const Matrix<double> &M, double ZeroThreshold) 
{
  Compute( M, ZeroThreshold);
}


int SVD::Compute(const Matrix<double> & M, double ZeroThreshold) 
{
  Decomposed_=false;
  ZeroThreshold_ = ZeroThreshold;
  return General_Eigenproblem_GeneralMatrix_Lapack( M );
}


int SVD::General_Eigenproblem_GeneralMatrix_Lapack(const Matrix<double> & M) 
{
  // S_ will contain the singular values of A sorted with s(i) <= s(i+1)  
  // U_ will contain m x m left orthogonal/unitary matrix with 
  //    eigenvectors in columns  
  // VT_ will contain n x n right _transposed_ orthogonal/unitary matrix 
  //     with eigenvectors in rows (because of transpose)    
  A_=M;
  int res=0;
  res = General_singular_value_decomposition(A_, S_, U_, VT_);
  if ( S_.size() != 0 && U_.num_rows() != 0 &&  U_.num_cols() != 0 &&
       VT_.num_rows() != 0 &&  VT_.num_cols() != 0) {
    Decomposed_=true;
  } else {
    res = -1;
  }
  
  return res;
}

// -- Solve the matrix-vector system M x = y, returning x.
Vector<double> SVD::Solve(const Vector<double>& y) const
{
  BIASERR("warning, there is a bug in this function, use Solve(Matrix, vector, Vector instead");
  Vector<double> x(VT_.num_cols());      // solution matrix vector
  Vector<double> UTy;

  //cout << "U_=" << U_ << " VT_= " << VT_ << " S_=" << S_;

  if(Decomposed_)    
    if (y.size() <= U_.num_rows()){
      Matrix<double> UT=U_.Transpose();
      // at first multiply with U^T
      if (U_.num_rows() < U_.num_cols()) {// augment y with extra rows of
        Vector<double> yy(U_.num_rows(), 0.0);      // zeros, so that it match
        for (int k=0; k< y.size();k++) 
          yy[k] = y[k]; // cols of u.transpose.
        UTy = UT * yy;
      } else {
        UTy = UT * y;
      }
           
      // multiply with diagonal 1/S_
      int i=0;
      for (; i < UTy.size(); i++)      
        x[i]=(fabs(S_[i]) >= ZeroThreshold_) ? UTy[i]/S_[i] : 0;//UTy[i]*S_[i];
      
      for (;i<x.size();i++)  x[i]=0;
      
      return VT_.Transpose() * x;   // premultiply with v.
      
    } else {
      BIASERR("size of right hand side of Ax = y is too big. Length(y) = "
              << y.size());
      return x;
    } else {
      BIASERR("SVD holds no decomposed matrix.");
      return x;
    }
}


Matrix<double> SVD::Invert()
{
  Matrix<double> V(VT_.num_cols(), VT_.num_rows());
  Matrix<double> UT(U_.num_cols(), U_.num_rows());
  Matrix<double> SdT(VT_.num_rows(), U_.num_cols());
  SdT.SetZero();
  for (register int i=0; i< S_.size(); i++){
    SdT[i][i] = (S_[i]<ZeroThreshold_ && S_[i]>-ZeroThreshold_)
      ?(0.0):(1/S_[i]);
  }
  V = VT_.Transpose();
  UT = U_.Transpose();

  if (Decomposed_)
    return V * SdT * UT;

  Matrix<double> res(0,0);
  return res;
}

Matrix<double> SVD::Invert(Matrix<double> A)
{
  if (Compute(A)==0)
    return Invert();
  else return Matrix<double>(0,0);
} 


const int SVD::NullspaceDim() const 
{
  int dim = 0;
  int index = S_.size()-1; 

  //cerr << "ZeroThreshold_: "<<ZeroThreshold_<<endl;
  while (index >= 0 && fabs(S_[index]) < ZeroThreshold_ ) {
    // singular values are sorted in descending order is still (around zero
    index--;
    dim++;
  }
  
  if (A_.num_cols()>A_.num_rows()){
      dim+=A_.num_cols()-A_.num_rows();
  }
  return dim;
}

const int SVD::RelNullspaceDim() const 
{ 
  int dim = 0;
  int index = S_.size()-1; 

  // use greatest singular value as reference
  const double RefSingValue(S_[0]);
  if (RefSingValue<=DBL_EPSILON) {
    // the greatest singular value is zero => all are zero
    return 0;
  } 
    
  while ((S_[index]/RefSingValue)<ZeroThreshold_){
    dim++;
    if (--index<1) break; // S[0]/S[0]==1 always => dont need to check that
  }

  // check if we have an "underdetermined equation system" by dimensionality
  if (A_.num_cols()>A_.num_rows()){
    dim += A_.num_cols()-A_.num_rows();
  }
   
  return dim;
}

const int SVD::RightNullspaceDim() const { 
  return NullspaceDim();
}

const int SVD::RelRightNullspaceDim() const { 
    return RelNullspaceDim();
}

const int SVD::LeftNullspaceDim() const 
{ 
  int dim = 0;
  int index = S_.size()-1; 

  //cerr << "ZeroThreshold_: "<<ZeroThreshold_<<endl;
  while (index >= 0 && fabs(S_[index]) < ZeroThreshold_ ) {
    // singular values are sorted in descending order is still (around zero
    index--;
    dim++;
  }
  
  if (A_.num_rows()>A_.num_cols()){
      dim+=A_.num_rows()-A_.num_cols();
  }
  return dim;
}

const int SVD::RelLeftNullspaceDim() const 
{ 
    int dim = 0;
    int index = S_.size()-1; 
  
    if (GetSingularValue(0)==0){
	dim=NullspaceDim();
    } else {
	double maxs=S_[0];
	while (index>=0 && fabs(S_[index]/maxs)<ZeroThreshold_){
	    index--;
	    dim++;
	}
	if (A_.num_rows()>A_.num_cols()){
	    dim+=A_.num_rows()-A_.num_cols();
	}
    } 
    return dim;
}

unsigned int SVD::Rank()
{
  BIASDOUT(D_SVD_RANK, "S_.Size(): "<<S_.Size()<<" NullspaceDim(): "<< NullspaceDim()<<"  A_.num_cols(): "<<A_.num_cols()<<"   A_.num_rows(): "<<A_.num_rows()<<"  S_: "<<S_);
  unsigned int lowdim=A_.num_rows()<A_.num_cols()?A_.num_rows():A_.num_cols();
  return (lowdim-NullspaceDim());
}


int SVD::Solve(Matrix<double>& A, Vector<double>& b, Vector<double>& x)
{
  Matrix<double> ATA(A.num_cols(), A.num_cols());
  ATA=A.Transpose()*A;
  Compute(ATA);
  //Compute(A);
  Vector<double> d(A.num_rows()), z(A.num_cols());
  Vector<double> S=GetS();
  BIASCDOUT(D_SVD_SOLVE, "S: "<<S<<endl); 
  d=GetVT() * A.Transpose() * b;
  //d=GetU().Transpose() * b;
  BIASCDOUT(D_SVD_SOLVE, "d: "<<d<<endl);
  for (int i=0; i<A.num_cols(); i++)
    z[i]=(fabs(S[i])>DEFAULT_DOUBLE_ZERO_THRESHOLD)?d[i]/S[i]:0.0;
  
  //x = GetV() * z;
  x = GetU() * z;
  return 0;
}

