/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdlib.h>
//#include <unistd.h>

#ifdef   WIN32
#include <Base/Common/getopt_W32.h>
#else  //not WIN32
#include <getopt.h>
#endif //WIN32

#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <time.h>
#include <algorithm>

#include "Param.hh"
#include "ParamTypes.hh"

// for umask
#include <sys/types.h>
#include <sys/stat.h>
#include <locale.h>

using namespace BIAS;
using namespace std;

// we set a typedef here, as we need const iterators very often...
typedef map<const std::string,BIAS::ParamTypeBase*>::const_iterator CI;


Param::Param()
{
  DestructorWarning_ = true;
  // init some members
  ParamMap_.clear();
#ifdef WIN32
  AddParamString("writeconfig","Write file with current configuration","",'w');
  AddParamString("readconfig","Read configuration from file","",'r');
#else
  AddParamString("writeconfig","Write file with current configuration");
  AddParamString("readconfig","Read configuration from file");
#endif
  SetParameterWriteToFile("writeconfig",false);
  SetParameterWriteToFile("readconfig",false);

  // needed to ensure consistent ,/. behavior in different envs
  setlocale(LC_ALL, "C");
}


Param::~Param()
{
  // destroy all listitems...
  for ( CI it=ParamMap_.begin(); it!=ParamMap_.end(); it++){
    // cerr << " deleting "<<it->first<<endl;
    delete it->second;
  }
#ifdef BIAS_DEBUG
  if (DestructorWarning_)
    BIASERR("\n Warning: Parameter are held in class Param. Don't delete this class, if you're still using the parameter somewhere.\n\n");
#endif
}


void Param::ShowData(std::ostream &os, int grp, bool showenhanced) 
{
  int namew=GetSizeOfLongestParamName()+2, typew=7;
  int valw=GetSizeOfLongestParamValueString()+2;
  
  namew=(namew<20)?(20):namew;
  valw=(valw<7)?(7):valw;
  
  int pos   = 0;
  
  CI it;

  os <<"idx"<<setw(namew)<<"name"<<setw(typew)
     <<"type"<<setw(valw)<<"value"<<" cmdl-opts"<<endl;

  if (grp==GRP_ALL){
    vector<int> grps=UniqueSortedGroups();
    vector<int>::iterator git;
    for (git=grps.begin(); git!=grps.end(); git++){
      os <<"\n############### Group ID "<<*git<<" : "<<GetGroupName(*git)
         <<" #######################\n";
      for(it=ParamMap_.begin(); it!=ParamMap_.end(); it++){
        if ((*it).second->Group_==*git){
          if (showenhanced || (!(*it).second->Enhanced_)) {
            os << pos << ": " << setw(namew) << it->first;
            os << setw(typew) << it->second->GetTypeName()<<" " 
               << setw(valw) << it->second->GetValueString() <<"   " 
               << it->second->ShortCmd_ ;
            if ((*it).second->Enhanced_) os << "   (enhanced)";
            os << endl;
            pos++;
          }
        }
      }
    }
  } else { 
    for(it=ParamMap_.begin(); it!=ParamMap_.end(); it++){
      if ((*it).second->Group_==grp){
        if (showenhanced || (!(*it).second->Enhanced_)) {
          os << pos << ": " << setw(namew) << it->first;
          os << setw(typew) << it->second->GetTypeName()<<" " 
             << setw(valw) << it->second->GetValueString() <<"   " 
             << it->second->ShortCmd_; 
          if ((*it).second->Enhanced_) os << "   (enhanced)";
          os << endl;
          pos++;
        }
      }
    }
  }
}


void Param::Usage(std::ostream &os) 
{
  int namew = 0;
  int typew = 9;
  char *tmpstr;

  // search the length of the largest name
  CI it;

  for(it=ParamMap_.begin(); it!=ParamMap_.end(); it++)
    {
      int len = it->first.length();
      if ( len>namew )
        namew = len;
    }
  
  // get space for "--" + name + 0
  tmpstr = new char[namew+3];
  
  os << endl;
  os << "Available Parameters with description and commandline switches";
  os << endl;

  vector<int> grps=UniqueSortedGroups();
  vector<int>::iterator git;
  for (git=grps.begin(); git!=grps.end(); git++){
    os <<"\n############### Group ID "<<*git<<" : "<<GetGroupName(*git)
        <<" #######################\n";
    for(it=ParamMap_.begin(); it!=ParamMap_.end(); it++){
      if (it->second->Group_ == *git && (!(*it).second->Enhanced_)){
        sprintf(tmpstr,"--%s",it->first.c_str());
        os <<setw(namew+2)<<tmpstr;

        if (it->second->ShortCmd_ == 0)
          os << "    ";
        else
          os << ", -" << it->second->ShortCmd_;    

        if ( ParamTypeInt *item =
             dynamic_cast<ParamTypeInt*>(it->second) )
          {
            // element is of type INTEGER
            os << setw(typew) << "INT ";
            item = NULL; // to fool compiler and suppress warning ;-)
          }
        else if ( ParamTypeBool *item = 
                  dynamic_cast<ParamTypeBool*>(it->second) )
          {
            // element is of type BOOLEAN
            os << setw(typew) << "BOOL ";
            item = NULL; // to fool compiler and suppress warning ;-)
          }
        else if ( ParamTypeDouble *item =
                  dynamic_cast<ParamTypeDouble*>(it->second) )
          {
            // element is of type DOUBLE
            os << setw(typew) << "DOUBLE ";
            item = NULL; // to fool compiler and suppress warning ;-)
          }
        else if ( ParamTypeString *item =
                  dynamic_cast<ParamTypeString*>(it->second) )
          {
            // element is of type STRING
            os << setw(typew) << "STRING ";
            item = NULL; // to fool compiler and suppress warning ;-)
          }
        else if ( ParamTypeVecDbl *item =
                  dynamic_cast<ParamTypeVecDbl*>(it->second) )
          {
            // element is of type VECTOR OF DOUBLE
            os << setw(typew) << "VECDBL ";
            item = NULL; // to fool compiler and suppress warning ;-)
          }
        else if ( ParamTypeVecInt *item =
                  dynamic_cast<ParamTypeVecInt*>(it->second) )
          {
            // element is of type VECTOR OF INTEGER
            os << setw(typew) << "VECINT ";
            item = NULL; // to fool compiler and suppress warning ;-)
          }
        else
          {
            os << "unknown ";
          }

        os <<" : ";
        if ( it->second->Help_ != "" ) os << it->second->Help_;
        else os << "No help available";
    
        os<<endl;
      }
    }
    os<<endl;
  }
  delete[] tmpstr;  
}

// ADDING routines ---------------------------------------------------

int * Param::AddParamInt( const std::string &name, 
                          const std::string &help, int deflt, 
                          int min, int max ,char cmdshort, int Group)
{
  if ( ParamMap_.find(name)!=ParamMap_.end() )
    {
      BIASERR("Cannot add parameter "<<name<<". Already in list!");
      return NULL;
    }
  ParamTypeInt * NewItem = new ParamTypeInt();
  NewItem->Name_        = name;
  NewItem->Help_        = help;
  NewItem->ShortCmd_    = cmdshort;
  NewItem->Group_       = Group;
  NewItem->WriteToFile_ = true;
  NewItem->value_       = deflt;
  NewItem->default_     = deflt;
  NewItem->min_         = min;
  NewItem->max_         = max;
  pair<string, ParamTypeBase*> NewPair(name,NewItem);
  ParamMap_.insert(ParamMap_.end(), NewPair );
  return &(NewItem->value_);;
}


bool * Param::AddParamBool( const std::string &name, const std::string &help,
                            bool deflt, char cmdshort, int Group )
{
  if ( ParamMap_.find(name)!=ParamMap_.end() )
    {
      BIASERR("Cannot add parameter "<<name<<". Already in list!");
      return NULL;
    }
  ParamTypeBool * NewItem = new ParamTypeBool();
  NewItem->Name_        = name;
  NewItem->Help_        = help;
  NewItem->ShortCmd_    = cmdshort;
  NewItem->Group_       = Group;
  NewItem->WriteToFile_ = true;
  NewItem->value_       = deflt;
  NewItem->default_     = deflt;
  pair<string, ParamTypeBase*> NewPair(name,NewItem);
  ParamMap_.insert(ParamMap_.end(), NewPair );
  return &(NewItem->value_);
}


std::string * Param::AddParamString( const std::string &name, 
                                     const std::string &help,
                                     std::string deflt, char cmdshort,
                                     int Group )
{
  if ( ParamMap_.find(name)!=ParamMap_.end() )
    {
      BIASERR("Cannot add parameter "<<name<<". Already in list!");
      return NULL;
    }
  ParamTypeString * NewItem = new ParamTypeString();
  NewItem->Name_        = name;
  NewItem->Help_        = help;
  NewItem->ShortCmd_    = cmdshort;
  NewItem->Group_       = Group;
  NewItem->WriteToFile_ = true;
  NewItem->value_       = deflt;
  NewItem->default_     = deflt;
  pair<string, ParamTypeBase*> NewPair(name,NewItem);
  ParamMap_.insert(ParamMap_.end(), NewPair );
  return &(NewItem->value_);;
}


double * Param::AddParamDouble( const std::string &name, 
                                const std::string &help,
                                double deflt, 
				double min, double max,
				char cmdshort, int Group)
{
  if ( ParamMap_.find(name)!=ParamMap_.end() )
    {
      BIASERR("Cannot add parameter "<<name<<". Already in list!");
      return NULL;
    }
  ParamTypeDouble * NewItem = new ParamTypeDouble();
  NewItem->Name_        = name;
  NewItem->Help_        = help;
  NewItem->ShortCmd_    = cmdshort;
  NewItem->Group_       = Group;
  NewItem->WriteToFile_ = true;
  NewItem->value_       = deflt;
  NewItem->default_     = deflt;
  NewItem->min_         = min;
  NewItem->max_         = max;
  pair<string, ParamTypeBase*> NewPair(name,NewItem);
  ParamMap_.insert(ParamMap_.end(), NewPair );
  return &(NewItem->value_);;
}


BIAS::Vector<double> * Param::AddParamVecDbl( const std::string &name, 
                                              const std::string &help,
                                              const Vector<double> &deflt,
                                              char cmdshort, int Group )
{
  if ( ParamMap_.find(name)!=ParamMap_.end() )
    {
      BIASERR("Cannot add parameter "<<name<<". Already in list!");
      return NULL;
    }
  ParamTypeVecDbl * NewItem = new ParamTypeVecDbl();
  NewItem->Name_        = name;
  NewItem->Help_        = help;
  NewItem->Group_       = Group;
  NewItem->ShortCmd_    = cmdshort;
  NewItem->WriteToFile_ = true;
  NewItem->value_       = deflt;
  NewItem->default_     = deflt;
  pair<string, ParamTypeBase*> NewPair(name,NewItem);
  ParamMap_.insert(ParamMap_.end(), NewPair );
  return &(NewItem->value_);;
}

BIAS::Vector<int> * Param::AddParamVecInt( const std::string &name, 
                                           const std::string &help,
                                           const Vector<int> &deflt,
                                           char cmdshort, int Group )
{
  if ( ParamMap_.find(name)!=ParamMap_.end() )
    {
      BIASERR("Cannot add parameter "<<name<<". Already in list!");
      return NULL;
    }
  ParamTypeVecInt * NewItem = new ParamTypeVecInt();
  NewItem->Name_        = name;
  NewItem->Help_        = help;
  NewItem->Group_       = Group;
  NewItem->ShortCmd_    = cmdshort;
  NewItem->WriteToFile_ = true;
  NewItem->value_       = deflt;
  NewItem->default_     = deflt;
  pair<string, ParamTypeBase*> NewPair(name,NewItem);
  ParamMap_.insert(ParamMap_.end(), NewPair );
  return &(NewItem->value_);;
}

int * Param::AddParamEnum( const std::string &name, 
                           const std::string &help,
                           const std::vector<std::string> &enums,
                           const int deflt, 
			   const std::vector<int> * IDs,
                           const char cmdshort, 
			   const int Group )
{
  if ( ParamMap_.find(name)!=ParamMap_.end() )
    {
      BIASERR("Cannot add parameter "<<name<<". Already in list!");
      return NULL;
    }
  if ( enums.empty() )
    {
      BIASERR("At least one item is needed for enum type");
      return NULL;
    }
  if ( IDs != NULL && IDs->size() != enums.size() )
    {
      BIASERR("You gave "<<enums.size()<< "enums but "<<IDs->size()
              <<" ID values! Both numbers must be equal!");
      return NULL;
    }
  ParamTypeEnum * NewItem = new ParamTypeEnum();
  NewItem->Name_        = name;
  NewItem->Help_        = help;
  NewItem->Group_       = Group;
  NewItem->ShortCmd_    = cmdshort;
  NewItem->WriteToFile_ = true;
  NewItem->value_       = deflt;
  NewItem->default_     = deflt;
  pair<string,int> NewEnumPair;
  int newID;
  for(unsigned int i=0; i<enums.size(); i++)
    {
      newID = IDs?(*IDs)[i]:i;
      if ( NewItem->Map_.find(enums[i]) != NewItem->Map_.end() ) {
          BIASERR("Given enum IDs must be unique!");
          delete NewItem;
          return NULL;
        }
      NewEnumPair = make_pair(enums[i],newID );
      NewItem->Map_.insert(NewItem->Map_.end(), NewEnumPair);
    }
//   if ( NewItem->Map_.find(deflt) == NewItem->Map_.end() )
//     {
//       BIASERR("The given default ID is not in the enum list!");
//       delete NewItem;
//       return NULL;
//     }
  pair<string, ParamTypeBase*> NewPair(name,NewItem);
  ParamMap_.insert(ParamMap_.end(), NewPair );
  return &(NewItem->value_);
}


//////////////  CheckParam() ///////////////

bool Param::CheckParam(const std::string &name)
{
  if ( ParamMap_.find(name) == ParamMap_.end() ) return false;
  else return true;
}

// GETTING routines --------------------------------------------------

int * Param::GetParamInt( const std::string &name )const
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() ){
    BIASERR("int parameter "<<name<<" not found!");
    //   BIASASSERT(false);
      return NULL;
    }
  else if ( ParamTypeInt *item = dynamic_cast<ParamTypeInt*>(it->second) )
    return &(item->value_);
  else{
      BIASERR("Parameter "<<name<<" is not of type INTEGER!");
      return NULL;
    }
}


bool * Param::GetParamBool( const std::string &name )const
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() ){
    BIASERR("bool parameter "<<name<<" not found!");
      return NULL;
    }
  else if ( ParamTypeBool *item = dynamic_cast<ParamTypeBool*>(it->second) )
    return &(item->value_);
  else{
      BIASERR("Parameter "<<name<<" is not of type BOOLEAN!");
      return NULL;
    }
}


std::string * Param::GetParamString( const std::string &name )const
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() ){
    BIASERR("string parameter "<<name<<" not found!");
      return NULL;
    }
  else if (ParamTypeString *item=dynamic_cast<ParamTypeString*>(it->second))
    return &(item->value_);
  else{
      BIASERR("Parameter "<<name<<" is not of type STRING!");
      return NULL;
    }
}


double * Param::GetParamDouble( const std::string &name )const
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() ){
    BIASERR("double parameter "<<name<<" not found!");
    //    BIASASSERT(false);
    return NULL;
  }
  else if (ParamTypeDouble *item=dynamic_cast<ParamTypeDouble*>(it->second))
    return &(item->value_);
  else{
      BIASERR("Parameter "<<name<<" is not of type DOUBLE!");
      return NULL;
    }
}


BIAS::Vector<double> * Param::GetParamVecDbl( const std::string &name )const
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() ){
    BIASERR("vecdbl parameter "<<name<<" not found!");
      return NULL;
    }
  else if (ParamTypeVecDbl *item=dynamic_cast<ParamTypeVecDbl*>(it->second))
    return &(item->value_);
  else{
      BIASERR("Parameter "<<name<<" is not of type VECTOR OF DOUBLE!");
      return NULL;
    }
}


BIAS::Vector<int> * Param::GetParamVecInt( const std::string &name )const
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() ){
    BIASERR("vecint parameter "<<name<<" not found!");
      return NULL;
    }
  else if (ParamTypeVecInt *item=dynamic_cast<ParamTypeVecInt*>(it->second))
    return &(item->value_);
  else{
      BIASERR("Parameter "<<name<<" is not of type VECTOR OF INTEGER!");
      return NULL;
    }
}


int * Param::GetParamEnum( const std::string &name )const
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() ){
    BIASERR("enum parameter "<<name<<" not found!");
      return NULL;
    }
  else if ( ParamTypeEnum *item = dynamic_cast<ParamTypeEnum*>(it->second) )
    return &(item->value_);
  else{
      BIASERR("Parameter "<<name<<" is not of type ENUM!");
      return NULL;
    }
}

// SETTING routines --------------------------------------------------

int * Param::SetParamInt( const std::string &name, const int &value )
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() )
    {
      BIASERR("Parameter not found!");
      return NULL;
    }
  if ( ParamTypeInt *item = dynamic_cast<ParamTypeInt*>(it->second) )
    {
      if ( value < item->min_ || value > item->max_ )
        {
          BIASERR("Valid values for "<<name<<" are between "
                  <<item->min_<<" and "<<item->max_<<"!");
          return NULL;   
        }
      item->value_ = value;
      return &(item->value_);
    }      
  else
    {
      BIASERR("Parameter "<<name<<" is not of type INTEGER!");
      return NULL;
    }
}


bool * Param::SetParamBool( const std::string &name, const bool &value )
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() )
    {
      BIASERR("Parameter not found!");
      return NULL;
    }
  if ( ParamTypeBool *item = dynamic_cast<ParamTypeBool*>(it->second) )
    {
      item->value_ = value;
      return &(item->value_);
    }      
  else
    {
      BIASERR("Parameter "<<name<<" is not of type BOOLEAN!");
      return NULL;
    }
}


std::string * Param::SetParamString( const std::string &name,
                                     const std::string &value )
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() )
    {
      BIASERR("Parameter not found!");
      return NULL;
    }
  if ( ParamTypeString *item = dynamic_cast<ParamTypeString*>(it->second) )
    {
      item->value_ = value;
      return &(item->value_);
    }      
  else
    {
      BIASERR("Parameter "<<name<<" is not of type STRING!");
      return NULL;
    }
}


double * Param::SetParamDouble( const std::string &name, const double &value )
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() )
    {
      BIASERR("Parameter not found!");
      return NULL;
    }
  if ( ParamTypeDouble *item = dynamic_cast<ParamTypeDouble*>(it->second) )
    {
      if ( value < item->min_ || value > item->max_ )
        {
          BIASERR("Valid values for "<<name<<" are between "
                  <<item->min_<<" and "<<item->max_<<"!");
          return NULL;   
        }
      item->value_ = value;
      return &(item->value_);
    }      
  else
    {
      BIASERR("Parameter "<<name<<" is not of type DOUBLE!");
      return NULL;
    }
}


BIAS::Vector<double> * Param::SetParamVecDbl( const std::string &name,
                                              const Vector<double> &value )
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() )
    {
      BIASERR("Parameter not found!");
      return NULL;
    }
  if ( ParamTypeVecDbl *item = dynamic_cast<ParamTypeVecDbl*>(it->second) )
    {
//       if ( value.size() != item->default_.size() )
//         {
//           BIASERR("New vector for "<<name<<" must be of size "
//                   <<item->default_.size()<<" but this one's is "
//                   <<value.size()<<"!");
//           return NULL;   
//         }
      item->value_ = value;
      return &(item->value_);
    }      
  else
    {
      BIASERR("Parameter "<<name<<" is not of type VECTOR OF DOUBLE!");
      return NULL;
    }
}


BIAS::Vector<int> * Param::SetParamVecInt( const std::string &name,
                                           const Vector<int> &value )
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() )
    {
      BIASERR("Parameter not found!");
      return NULL;
    }
  if ( ParamTypeVecInt *item = dynamic_cast<ParamTypeVecInt*>(it->second) )
    {
//       if ( value.size() != item->default_.size() )
//         {
//           BIASERR("New vector for "<<name<<" must be of size "
//                   <<item->default_.size()<<" but this one's is "
//                   <<value.size()<<"!");
//           return NULL;   
//         }
      item->value_ = value;
      return &(item->value_);
    }      
  else
    {
      BIASERR("Parameter "<<name<<" is not of type VECTOR OF INTEGER!");
      return NULL;
    }
}


int * Param::SetParamEnum( const std::string &name,const std::string &value )
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() )
    {
      BIASERR("Parameter not found!");
      return NULL;
    }
  if ( ParamTypeEnum *item = dynamic_cast<ParamTypeEnum*>(it->second) )
    {
      map<string,int>::iterator it;
      if ( (it=item->Map_.find(value)) == item->Map_.end() ) {
          BIASERR(value << " is not a valid ID for this enum!");
          return NULL;
        }
      item->value_ = it->second;
      return &(item->value_);
    }      
  else
    {
      BIASERR("Parameter "<<name<<" is not of type ENUM!");
      return NULL;
    }
}


int * Param::SetParamEnum( const std::string &name, const int &value )
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() )
    {
      BIASERR("Parameter not found!");
      return NULL;
    }
  if ( ParamTypeEnum *item = dynamic_cast<ParamTypeEnum*>(it->second) ) {
    // rush through map and if value is valid, set it
    map<string, int>::iterator it;
    for(it=item->Map_.begin(); it!=item->Map_.end(); it++)
      if ( it->second == value ){
	item->value_ = value;
	return &(item->value_);
      }
    BIASERR(value << " is not in this enum!");
    return NULL;
  }      
  else {
    BIASERR("Parameter "<<name<<" is not of type ENUM!");
    return NULL;
  }
}


int Param::SetParameterWriteToFile(const std::string &name, bool writeToFile)
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() )
    {
      BIASERR("Parameter not found!");
      return -1;
    }
  it->second->WriteToFile_ = writeToFile;
  return 0;
}

int Param::SetEnhancedFlag(const std::string &name, bool enhanced)
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() )
    {
      BIASERR("Parameter not found! "<<name);
      return -1;
    }
  it->second->Enhanced_ = enhanced;
  return 0;
}


// command line and file processing methods --------------------------

int Param::UpdateParameter(int &argc, char *argv[], 
                           const std::string &default_filename)
{
  int res;
  // provide a default file name to readconfig
  *GetParamString("readconfig")=default_filename;
  res = ParseCommandLine(argc, argv);
  string *name=GetParamString("writeconfig");

  if (*name!="") {
    //cerr << "*name "<<*name<<endl;
    WriteParameter(*name, true);
    BIASCDOUT(D_PARAM_CONF_FILE, "wrote default param file "
              <<default_filename<<endl);
  } else {
    // write default parameter file if no parameter file given and default
    // file not found
    if (*GetParamString("readconfig")==default_filename) {
      BIASCDOUT(D_PARAM_CONF_FILE, "using default param file "
                <<default_filename<<endl);
      ifstream is(default_filename.c_str());
      if (!is) {
        is.close();
        WriteParameter(default_filename, false);
        BIASCDOUT(D_PARAM_CONF_FILE, "wrote default param file "
                  <<default_filename<<endl);
      } else {
        is.close();
        BIASCDOUT(D_PARAM_CONF_FILE, "found default param file "
                  <<default_filename<<endl);
      }
    } else {
      BIASCDOUT(D_PARAM_CONF_FILE, "using user given param file "
                <<*GetParamString("readconfig")<<endl);
    }
  }
  return res;
}

int Param::ParseCommandLine(int &argc, char *argv[])
{
  // if no arguments are given on Cmdl, return index to last non-opt: 1
  string *readPrj;
  if (argc==1) {
    // read configuration file if one is given (see UpdateParameter)
    if ( (readPrj = GetParamString("readconfig"))!=NULL ){
      if (*readPrj != ""){
        //cout <<"Reading configuration from file: "<<*readPrj<<endl;
        BIASCDOUT(D_PARAM_CONF_FILE, "Reading configuration from file: "
                  <<*readPrj<<endl);
        if (ReadParameter(*readPrj)<0) return -1;
      }
    }
    return 1;
  }
  BIASDOUT(D_PARAM_PARSECMDL,"Param::ParseCommandLine()");

  // fisrt get a copy of all that if we not already have one
  if (ArgV_.size() ==0)
    for ( int i=0;i<argc; i++)
      ArgV_.push_back(argv[i]);

  //BIASDOUT(D_PARAM_PARSECMDL,"NumParams_: "<<ParamMap_.size());  
  int itemcount = 0;
  // there are NumParams_ shortopts maximum, each may have one arg,
  // plus trailing 0
  char *shortoptstring=new char[ParamMap_.size()*3+1];
  bool bReadConfig =true, bParse=true;
  while (bParse) {
    // set string empty
    shortoptstring[0] = 0;
    itemcount = 0;
    char tmpstr[2];
    tmpstr[1] = 0;

    /** @todo WIN32 does not support getopt_long(), maybe we could use 
	posix-runtime environment for WIN32  (evers 2004-10-14) */
#ifndef WIN32
    // prepare array of longopts for getopt_long()
    // last element must be zeored
    longopts = new struct option[ParamMap_.size()+1];
    memset(&longopts[ParamMap_.size()], 0, sizeof(struct option));
#endif //WIN32
  
    // traverse list of parameters and fill in longopts[]

    for (CI it=ParamMap_.begin(); it!=ParamMap_.end(); ++it)
      {
        //getopt_long not present in WIN32
#ifndef WIN32
        longopts[itemcount].name = (it->first).c_str();
        // only type BOOL has an optional argument
        if ( dynamic_cast<ParamTypeBool*>(it->second) != NULL )
          longopts[itemcount].has_arg = optional_argument;
        else longopts[itemcount].has_arg = required_argument;

        // set .flag to NULL, then .val is returned from getopt_long()
        longopts[itemcount].flag = NULL;
        // set .val to itemcount+offset let us find the target in ParamList_ 
        // much faster. Offset is required to differ from shortopt char
        longopts[itemcount].val = itemcount+1000;
#endif //WIN32

        // prepare shortoptstring
        if ( it->second->ShortCmd_!=0 ) {
          tmpstr[0] = it->second->ShortCmd_;
          strcat(shortoptstring, tmpstr);
          // append ':' to shortopstring for arguments to options
          strcat(shortoptstring, ":");
          // append a second ':' to bool, for optional argument
          if ( dynamic_cast<ParamTypeBool*>(it->second) != NULL)
            strcat(shortoptstring, ":");
        }
        itemcount++;
      }
#ifndef WIN32
    // reset optind to start scanning from 0
    optind = 0;
#endif

    bool done = false;
    int short_or_index;
    // now scan
    while (!done)
      {
        // getopt_long() returns a char if the short option was found
        // if the long option was found, the longopts.val is returned
        // in this case the index to longopts[] + 1000
        //getopt_long not present in WIN32
#ifndef WIN32
        short_or_index = getopt_long(argc, argv, shortoptstring, 
                                     longopts, NULL);
#else //WIN32
        short_or_index = getopt(argc, argv, shortoptstring);
#endif //WIN32

        // '?' means unknown option, ':' means missing argument
        if (short_or_index == '?' || short_or_index == ':') {
#ifndef  WIN32
          delete [] longopts;
#endif //WIN32
          Usage();
          return -1;
        }

        if (short_or_index == -1)
          done = true;
        else
          {
            // decide if short or long found
            CI it;
            if (short_or_index >=1000){
              // long opt found, take name from longopts and search in map
              int index = short_or_index -1000;
#ifndef  WIN32
              it = ParamMap_.find(longopts[index].name);
              BIASDOUT(D_PARAM_PARSECMDL,"longopt at "<<index
                       <<" : "<<it->first);
#endif //WIN32
            }
            else{
              for(it=ParamMap_.begin();it!=ParamMap_.end(); it++)
                if ( it->second->ShortCmd_==short_or_index ){
                  BIASDOUT(D_PARAM_PARSECMDL,"shortopt with "
                           <<(char)short_or_index<<" : "<<it->first);
                  break;
                }
            }
            // if optarg ==NUL, user specified a bool switch and no argument
            // or argument is not connected with '=', so fake argument
            if (optarg == NULL) optarg="true";
            // this calls the method from the correct class to insert into
            // (late binding magic)
            it->second->SetFromString(optarg);
          } // else if (short_or_index == -1)	  
      } //while ()
    /// this delete let the program bail out. but it realy should be deleted
#ifndef  WIN32
    delete[] longopts;
#endif //WIN32

    //// check for readconfig
    if (bReadConfig){    
      if ( (readPrj = GetParamString("readconfig"))!=NULL ){
        if (*readPrj != ""){
          cout <<"Reading configuration from file: "<<*readPrj<<endl;
          BIASCDOUT(D_PARAM_CONF_FILE, "Reading configuration from file: "
                    <<*readPrj<<endl);
          if (ReadParameter(*readPrj)<0) return -1;
          // second parse to let cmdl override config file
          else bParse=true;
        } else {
          bParse=false;
          BIASCDOUT(D_PARAM_CONF_FILE, "empty configuration file name\n");
        }
      } else { 
        bParse=false;
        BIASCDOUT(D_PARAM_CONF_FILE, "no configuration file given\n");
      }
      bReadConfig=false;
    } else {
      bParse=false;
    }
  } // end while bParse

  string *writePrj;

  // now eventually write config
  if ( (writePrj=GetParamString("writeconfig"))!=NULL ) {
    if (*writePrj != "") {
      string s = *writePrj;
      SetParamString("writeconfig",(string)"");
      SetParamString("readconfig",(string)"");
      WriteParameter(s);
    }
  }
  delete[] shortoptstring;
  return optind;
}//////////////// End of ParseCommandLine()


#define READPARAM_CHECKDUPES
int Param::ReadParameter(const string &filename)
{
  ifstream ifs;
  string typestr;
  string namestr;
  string assignstr;
  string valstr;
  char commentstr[1000];
  CI it; // const iterator for our ParamMap_

#ifdef READPARAM_CHECKDUPES
  map<string,bool> *ExistingNames = new map<string,bool>;
  int line = 0;
  string current;
#endif


  ifs.open(filename.c_str());
  if (!ifs) {
    BIASERR("error reading parameter file: "<<filename);
    ifs.close();
    return -1;
  }
  while (ifs) {

#ifdef READPARAM_CHECKDUPES
    line++;
#endif
  
    // try read one word, may be type or comment sign
    ifs >>  namestr;
    if (namestr[0] == '#') {
      ifs.getline(commentstr,sizeof(commentstr),'\n');
      BIASDOUT(D_PARAM_IO,"Comment read: "<<namestr<<commentstr);
      
    }
    else {
      ifs >> assignstr >>valstr ;
      // detect strings with whitespace and read them 
      if (valstr[0] == '"' && valstr[valstr.length()-1] != '"') {
	char c = 0;
	while ( c != '\"') { 
	  ifs.read(&c,1);
	  valstr.append(1,c);
	}
      }

      ifs >> typestr;
      if (ifs) {
// 	cout <<"read: "<<typestr<<" "<<namestr<<" "<<assignstr<<" "<<valstr
// 	     <<endl;
#ifdef READPARAM_CHECKDUPES
        current = (string)namestr;
        //cout << "checking line " << line << " name "<<current<<endl;
        if ( ExistingNames->find(current) != ExistingNames->end() ){
          cerr <<"Oops! Found duplicate parameter '" << current
               << "' in line " << line << " of file '" << filename 
               << "' " << endl;
          ifs.close();
          return -2;
        }
        else{
          pair<string, bool> NewPair(current,true);
          ExistingNames->insert(ExistingNames->end(), NewPair );
        };

#endif
	// remove double-quotes where necessary
        if (valstr[0] == '"' && valstr[valstr.length()-1] == '"') {
	  valstr = valstr.substr(1,valstr.length()-2);
        }
// 	cout <<"read: "<<typestr<<" "<<namestr<<" "<<assignstr<<" "<<valstr
// 	     <<endl;

        if ( (it=ParamMap_.find(namestr))==ParamMap_.end() ){
          BIASERR("Unknown parameter \""<<namestr
                  <<"\" in file "<<filename<<endl);
//           ShowData();
//           BIASASSERT(false);
        }
        else {
          // this calls the method from the correct class to insert
          // to (late binding magic)
          it->second->SetFromString(valstr);
          BIASDOUT(D_PARAM_IO, filename<<" parameter: "<<namestr
                   <<" type: "<<typestr << " valstr: "<<valstr);
        } // if (item == NULL) else {
      } // if(ifs)
    } // if (comment) else {
  } // While (ifs)  
  ifs.close();
  return 0;
}

int Param::GetFreeGroupID()
{
  vector<int> gv=UniqueSortedGroups();
  if (gv.size()==0){
    return 0;
  } 
  return (*(gv.rbegin()))+1;
}

int Param::GetGroupID(const string &name)
{
  CI it = ParamMap_.find(name); 	   
  if (it != ParamMap_.end()){ 	    
    return it->second->Group_; 	 
  }  
  BIASERR("no parameter with name "<<name); 	 
  return GRP_ALL;	     
}

int Param::GetGroupIDForGroupName(const string &name)
{
  for (unsigned int i=0;i<GroupNames_.size();i++)
    if (GroupNames_[i] == name) return i;
  return GRP_ALL;
}


int Param::SetGroupName(const int id, const string& name)
{
  int res=0;
  vector<int> ids=UniqueSortedGroups();
  if (find(ids.begin(), ids.end(), id)==ids.end()){
    BIASERR("unused group id "<<id);
    BIASASSERT(false);
    res=-1;
  } else {
    if (((*ids.rbegin())+1)>(int)GroupNames_.size()){
      GroupNames_.resize((*ids.rbegin())+1, "");
    }
    GroupNames_[id]=name;
  }
  return res;
}

std::string Param::GetGroupName(const int group_id)
{
  if (group_id==GRP_ALL){
    return "all";
  }
  if (group_id==GRP_NOSHOW){
    return "no show";
  }
  if (group_id>(int)(GroupNames_.size())-1 || group_id<0){
    BIASERR("invalid group id or no name given");
    //BIASASSERT(false);
    return "";
  }
  return GroupNames_[group_id];
}


vector<int> Param::UniqueSortedGroups()
{
  vector<int> res;
  CI it;
  for (it=ParamMap_.begin(); it!=ParamMap_.end(); it++){
    if (it->second->WriteToFile_) res.push_back(it->second->Group_);
  }
  sort(res.begin(), res.end());
  vector<int>::iterator end = unique(res.begin(), res.end());
  res.erase(end, res.end());
  return res;
}

bool Param::IsUsedGroupID(const int group_id)
{
  vector<int> ids=UniqueSortedGroups();
  return (find(ids.begin(), ids.end(), group_id)!=ids.end());
}

int Param::GetSizeOfLongestParamName()
{
  unsigned length=0;
  // search for longest name 
  for (CI it=ParamMap_.begin(); it!=ParamMap_.end(); it++){
    if (it->second->Name_.size()>length &&
        !it->second->Enhanced_)
      length=it->second->Name_.size();
  }
  return (int)length;
}

int Param::GetSizeOfLongestParamValueString()
{
  unsigned length=0;
  // search for longest name 
  for (CI it=ParamMap_.begin(); it!=ParamMap_.end(); it++){
    if (it->second->GetValueString().size()>length &&
        !it->second->Enhanced_)
      length=it->second->GetValueString().size();
  }
  return (int)length;
}


int Param::WriteParameter(const string &filename, bool write_comments, 
                          bool show_hidden)
{
  ofstream ofs;
  // search for longest name to set namew
  int namew=GetSizeOfLongestParamName()+2, typew=12;
  int valuew=GetSizeOfLongestParamValueString()+2;
  
  namew=(namew<20)?(20):namew;
  valuew=(valuew<7)?(7):valuew;
  
  ofs.open(filename.c_str());
  if ( !ofs.is_open() )
    BIASERR( "File '" << filename << "' not opened!" << endl);

  ostringstream oss;
  vector<int> grps=UniqueSortedGroups();
  vector<int>::iterator git;
  for (git=grps.begin(); git!=grps.end(); git++){
    ofs <<"\n############### Group ID "<<*git<<" : "<<GetGroupName(*git)
        <<" #######################\n";
    for ( CI it=ParamMap_.begin(); it!=ParamMap_.end(); it++){
      if (it->second->WriteToFile_ && it->second->Group_ == *git 
          && (show_hidden || (!(*it).second->Enhanced_)))
        {    
          oss.str("");
          if (write_comments)
            // first, write name with comment
            ofs <<endl<<"# "<<it->second->Name_<<":  "<<it->second->GetHint()
                <<endl<<"# "<<it->second->Help_<<endl;
          // second write name = value
          oss <<setw(namew)<<it->first<<" = "
              <<setw(valuew)<<left<<it->second->GetValueString()<<right
              <<"  " ;
          while (oss.str().size() <40) oss<<" ";
          ofs <<oss.str();
          oss.str("");
          // third write type name
          oss << "PARAM_"<<it->second->GetTypeName();
          ofs << setw(typew) << oss.str()<< endl;
        }
    }
  }
  ofs.close();
  return 0;
}

int  Param::WriteRunMe(const std::string &filename)
{
  ofstream ofs;
  time_t now_s;
  struct tm* now;
  char datestr[20];
  string fname;

  time(&now_s);
  now = localtime(&now_s);//, &now);
  sprintf(datestr,"%4d-%02d-%02d-%02d:%02d",now->tm_year+1900,now->tm_mon+1,now->tm_mday,
	  now->tm_hour,now->tm_min);
  
  fname = filename + datestr;
  ofs.open(fname.c_str());
 
  if (!ofs){
    perror(fname.c_str());
    return -1;
  }
  ofs <<"#!/bin/sh"<<endl<<endl;
  for (unsigned int i=0;i<ArgV_.size(); i++)
    ofs <<ArgV_[i]<<" ";
  ofs<<endl<<endl;
  ofs <<AllCvsDates<<endl;
  ofs.close();
#ifndef  WIN32
  chmod(filename.c_str(),0755);
#endif //WIN32
  return 0;
}



void Param::SetDefaults()
{
  for (CI it=ParamMap_.begin(); it!=ParamMap_.end(); it++)
    it->second->SetDefault();
}


int Param::RegisterCallbackObject(const string &name,
				  ParamCallbackInterface *obj)
{
  CI it; 
  if ( (it=ParamMap_.find(name)) != ParamMap_.end()) {
    it->second->CallbackObject_ = obj;
    return 0;
  }
  BIASERR("Unknown parameter \""<<name<<"\"in RegisterCallbackObject()");
  return -1;
}

int  Param::SetRangeInt(const std::string &name,const int min, const int max)
{
  CI it; 
  if ( (it=ParamMap_.find(name)) != ParamMap_.end()) {
    ParamTypeInt *pInt = dynamic_cast<ParamTypeInt *>(it->second);
    if (pInt != NULL) {
      pInt->min_ = min;
      pInt->max_  = max;
    }
    return 0;
  }
  BIASERR("Unknown parameter \""<<name<<"\"in RegisterCallbackObject()");
  return -1;
}


int Param::SetShortCommand(const string &name, const char cmd)
{
  CI it = ParamMap_.find(name);
  if ( it==ParamMap_.end() ){
    BIASERR("parameter "<<name<<" not found!");
      return -1;
    }
  else 
    it->second->ShortCmd_ = cmd;
  return 0;
}

